from ._base import WorldObject
from ..utils import unpack_bitfield
from .. import linalg


class Group(WorldObject):
    """A group of objects.

    A Group is useful when manipulating the scene graph as children can be
    jointly moved/scaled/rotated. It has no visual properties.

    Parameters
    ----------
    visible : bool
        If true, the object and its children are visible.
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """

    def __init__(self, *, visible=True, position=None):
        super().__init__(
            visible=visible,
            position=position,
        )


class Scene(Group):
    """Root of the scene graph.

    The scene holds scene-level information (background color, fog, environment
    map) as well as all objects that take part in the rendering process as
    either direct or indirect children/nested objects.

    """

    def __init__(self):
        super().__init__()


class Background(WorldObject):
    """The scene's background.

    Can be e.g. a gradient, a static image or a skybox.

    Parameters
    ----------
    geometry : Geometry
        Must be ``None``. Exists for compliance with the generic WorldObject
        API.
    material : Material
        The material to use when rendering the background.
    kwargs : Any
        Additional kwargs are forwarded to the object's :class:`base class
        <pygfx.objects.WorldObject>`.

    """

    def __init__(self, geometry=None, material=None, **kwargs):
        if geometry is not None and material is None:
            raise TypeError("You need to instantiate using Background(None, material)")
        super().__init__(None, material, **kwargs)


class Line(WorldObject):
    """An object representing a line using a list of vertices (3D positions).

    Some materials will render the line as a continuous line, while other materials
    will consider each pair of consequtive points a segment.

    The picking info of a Line (the result of ``renderer.get_pick_info()``) will
    for most materials include ``vertex_index`` (int) and ``segment_coord``
    (float, sub-segment coordinate).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object.
    material : Material
        The data defining the appearence of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        The render order (when applicable for the renderer's blend mode).
    render_mask : str
        Determines the render passes that the object is rendered in. It's
        recommended to let the renderer decide, using "auto".
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """


class Points(WorldObject):
    """A point cloud.

    An object consisting of points represented by vertices (3D positions).

    The picking info of a Points object (the result of
    ``renderer.get_pick_info()``) will for most materials include
    ``vertex_index`` (int) and ``point_coord`` (tuple of 2 float coordinates in
    logical pixels).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object.
    material : Material
        The data defining the appearence of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        The render order (when applicable for the renderer's blend mode).
    render_mask : str
        Determines the render passes that the object is rendered in. It's
        recommended to let the renderer decide, using "auto".
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """


class Mesh(WorldObject):
    """A mesh.

    An object consisting of triangular faces represented by a set of vertices
    (3D positions) and a set of vertex indices indicating which vertex triplets
    form mesh triangles.

    The picking info of a Mesh (the result of ``renderer.get_pick_info()``) will
    for most materials include ``instance_index`` (int), ``face_index`` (int),
    and ``face_coord`` (tuple of 3 floats). The latter are the barycentric
    coordinates for each vertex of the face (with values 0..1).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object.
    material : Material
        The data defining the appearence of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        The render order (when applicable for the renderer's blend mode).
    render_mask : str
        Determines the render passes that the object is rendered in. It's
        recommended to let the renderer decide, using "auto".
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """


class Image(WorldObject):
    """A 2D image.

    The geometry for this object consists only of ``geometry.grid``: a
    texture with the 2D data.

    If no colormap is applied to the material, the data are interpreted as
    colors in sRGB space. To use physical space instead, set the texture's
    colorspace property to ``"physical"``.

    The picking info of an Image (the result of ``renderer.get_pick_info()``)
    will for most materials include ``index`` (tuple of 2 int), and
    ``pixel_coord`` (tuple of float subpixel coordinates).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object.
    material : Material
        The data defining the appearence of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        The render order (when applicable for the renderer's blend mode).
    render_mask : str
        Determines the render passes that the object is rendered in. It's
        recommended to let the renderer decide, using "auto".
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """

    def _wgpu_get_pick_info(self, pick_value):
        tex = self.geometry.grid
        if hasattr(tex, "texture"):
            tex = tex.texture  # tex was a view
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, x=22, y=22)
        x = values["x"] / 4194304 * tex.size[0] - 0.5
        y = values["y"] / 4194304 * tex.size[1] - 0.5
        ix, iy = int(x + 0.5), int(y + 0.5)
        return {
            "index": (ix, iy),
            "pixel_coord": (x - ix, y - iy),
        }


class Volume(WorldObject):
    """A 3D image.

    The geometry for this object consists only of ``geometry.grid``: a texture
    with the 3D data.

    The picking info of a Volume (the result of ``renderer.get_pick_info()``)
    will for most materials include ``index`` (tuple of 3 int), and
    ``voxel_coord`` (tuple of float subpixel coordinates).

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object.
    material : Material
        The data defining the appearence of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        The render order (when applicable for the renderer's blend mode).
    render_mask : str
        Determines the render passes that the object is rendered in. It's
        recommended to let the renderer decide, using "auto".
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """

    def _wgpu_get_pick_info(self, pick_value):
        tex = self.geometry.grid
        if hasattr(tex, "texture"):
            tex = tex.texture  # tex was a view
        # This should match with the shader
        values = unpack_bitfield(pick_value, wobject_id=20, x=14, y=14, z=14)
        texcoords_encoded = values["x"], values["y"], values["z"]
        size = tex.size
        x, y, z = [(v / 16384) * s - 0.5 for v, s in zip(texcoords_encoded, size)]
        ix, iy, iz = int(x + 0.5), int(y + 0.5), int(z + 0.5)
        return {
            "index": (ix, iy, iz),
            "voxel_coord": (x - ix, y - iy, z - iz),
        }


class Text(WorldObject):
    """A text.

    See :class:``pygfx.TextGeometry`` for details.

    Parameters
    ----------
    geometry : TextGeometry
        The data defining the glyphs that make up the text.
    material : Material
        The data defining the appearence of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        The render order (when applicable for the renderer's blend mode).
    render_mask : str
        Determines the render passes that the object is rendered in. It's
        recommended to let the renderer decide, using "auto".
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    """

    uniform_type = dict(
        rot_scale_transform="4x4xf4",
    )

    def update_matrix(self, *args, **kwargs):
        super().update_matrix(*args, **kwargs)
        # When rendering in screen space, the world transform is used
        # to establish the point in the scene where the text is placed.
        # The only part of the local transform that is used is the
        # position. Therefore, we also keep a transform containing the
        # local rotation and scale, so that these can be applied to the
        # text in screen coordinates.
        matrix = linalg.Matrix4()
        matrix.compose(linalg.Vector3(0, 0, 0), self.rotation, self.scale)
        self.uniform_buffer.data["rot_scale_transform"].flat = matrix.elements



import numpy as np

from ._buffer import Resource, STRUCT_FORMAT_ALIASES


class Texture(Resource):
    """Container for textures.

    A base texture wrapper that can be implemented for numpy, ctypes arrays, or
    any other kind of array.

    Parameters:
        data : array, optional
            Array data of any type that supports the buffer-protocol, (e.g. a
            bytes or numpy array). If None, nbytes and nitems must be provided.
            The data is copied if it's not float32 or not contiguous.
        dim : int
            The dimensionality of the array (1, 2 or 3).
        size : tuple, [3]
            The extent ``(width, height, depth)`` of the array. If None, it is
            derived from `dim` and the shape of the data. The texture can also
            represent a stack of images by setting `dim=2` and `depth > 1`. Any
            derived texture views must then have a `view_dim` of  either
            'd2_array' or 'cube'.
        format : str
            A format string describing the texture layout. If None, this is
            automatically set from the data. This must be a pygfx format
            specifier, e.g. "3xf4", but can also be a format specific to the
            render backend if necessary (e.g. from ``wgpu.VertexFormat``).
        colorspace : str
            If this data is used as color, it is interpreted to be in this
            colorspace. Can be "srgb" or "physical".
        generate_mipmaps : bool
            If True, automatically generates mipmaps when transfering data to
            the GPU.

    """

    def __init__(
        self,
        data=None,
        *,
        dim,
        size=None,
        format=None,
        colorspace="srgb",
        generate_mipmaps=False,
    ):
        super().__init__()
        self._rev = 0
        # The dim specifies the texture dimension
        assert dim in (1, 2, 3)
        self._store.dim = int(dim)
        # The actual data (optional)
        self._data = None
        self._pending_uploads = []  # list of (offset, size) tuples
        self._mip_level_count = 1

        # Backends-specific attributes for internal use
        self._wgpu_usage = 0

        self._store.format = None if format is None else str(format)

        self._colorspace = (colorspace or "srgb").lower()
        assert self._colorspace in ("srgb", "physical")

        self._generate_mipmaps = bool(generate_mipmaps)

        size = None if size is None else (int(size[0]), int(size[1]), int(size[2]))

        if data is not None:
            mem = memoryview(data)
            if mem.format == "d":
                raise ValueError("Float64 data is not supported, use float32 instead.")
            self._data = data
            self._mem = mem
            self._store.nbytes = mem.nbytes
            self._store.size = self._size_from_data(mem, dim, size)
            self.update_range((0, 0, 0), self.size)
        elif size is not None and format is not None:
            self._store.size = size
            self._store.nbytes = 0
        else:
            raise ValueError(
                "Texture must be instantiated with either data or size and format."
            )

    @property
    def rev(self):
        """An integer that is increased when update_range() is called."""
        return self._rev

    @property
    def colorspace(self):
        """If this data is used as color, it is interpreted to be in this colorspace.
        Can be "srgb" or "physical". Default "srgb".
        """
        return self._colorspace

    @property
    def generate_mipmaps(self):
        """Whether to automatically generate mipmaps when uploading to the GPU."""
        return self._generate_mipmaps

    def get_view(self, **kwargs):
        """Get a new view on the this texture."""
        return TextureView(self, **kwargs)

    @property
    def dim(self):
        """The dimensionality of the texture (1, 2, or 3)."""
        return self._store.dim

    @property
    def data(self):
        """The data for this texture. Can be None if the data only
        exists on the GPU.

        Note: the data is the same reference that was given to
        instantiate this object, but this may change.
        """
        return self._data

    @property
    def mem(self):
        """The data for this buffer as a memoryview. Can be None if
        the data only exists on the GPU.
        """
        return self._mem

    @property
    def nbytes(self):
        """Get the number of bytes in the texture."""
        return self._store.nbytes

    @property
    def size(self):
        """The size of the texture as (width, height, depth).
        (always a 3-tuple, regardless of the dimension).
        """
        return self._store.size

    @property
    def format(self):
        """The texture format as a string. Usually a pygfx format specifier
        (e.g. u2 for scalar uint16, or 3xf4 for RGB float32),
        but can also be a overriden to a backend-specific format.
        """
        format = self._store.format
        if format is not None:
            return format
        elif self.data is not None:
            self._store["format"] = format_from_memoryview(self.mem, self.size)
            return self._store.format
        else:
            raise ValueError("Texture has no data nor format.")

    def update_range(self, offset, size):
        """Mark a certain range of the data for upload to the GPU.
        The offset and (sub) size should be (width, height, depth)
        tuples. Numpy users beware that an arrays shape is (height, width)!
        """
        # Check input
        assert isinstance(offset, tuple) and len(offset) == 3
        assert isinstance(size, tuple) and len(size) == 3
        if any(s == 0 for s in size):
            return
        elif any(s < 0 for s in size):
            raise ValueError("Update size must not be negative")
        elif any(b < 0 for b in offset):
            raise ValueError("Update offset must not be negative")
        elif any(b + s > refsize for b, s, refsize in zip(offset, size, self.size)):
            raise ValueError("Update size out of range")
        # Apply - consider that texture arrays want to be uploaded per-texture
        # todo: avoid duplicates by merging with existing pending uploads
        if self.dim == 1:
            for z in range(size[2]):
                for y in range(size[1]):
                    offset2 = offset[0], y, z
                    size2 = size[0], 1, 1
                    self._pending_uploads.append((offset2, size2))
        elif self.dim == 2:
            for z in range(size[2]):
                offset2 = offset[0], offset[1], z
                size2 = size[0], size[1], 1
                self._pending_uploads.append((offset2, size2))
        else:
            self._pending_uploads.append((offset, size))
        self._rev += 1

    def _size_from_data(self, data, dim, size):
        # Check if shape matches dimension
        shape = data.shape

        if size:
            # Get version of size with trailing ones stripped
            size2 = size
            size2 = size2[:-1] if size2[-1] == 1 else size2
            size2 = size2[:-1] if size2[-1] == 1 else size2
            rsize = tuple(reversed(size2))
            # Check if size matches shape
            if rsize != shape[: len(rsize)]:
                raise ValueError(f"Given size does not match the data shape.")
            return size
        else:
            if len(shape) not in (dim, dim + 1):
                raise ValueError(
                    f"Can't map shape {shape} on {dim}D tex. Maybe also specify size?"
                )
            # Determine size based on dim and shape
            if dim == 1:
                return shape[0], 1, 1
            elif dim == 2:
                return shape[1], shape[0], 1
            else:  # dim == 3:
                return shape[2], shape[1], shape[0]

    def _get_subdata(self, offset, size, pixel_padding=None):
        """Return subdata as a contiguous array."""
        # If this is a full range, this is easy
        if offset == 0 and size == self.nitems and self.mem.contiguous:
            return self.mem
        # Get a numpy array, because memoryviews do not support nd slicing
        if isinstance(self.data, np.ndarray):
            arr = self.data
        elif not self.mem.c_contiguous:
            raise ValueError(
                "Non-contiguous texture data is only supported for numpy array."
            )
        else:
            arr = np.frombuffer(self.mem, self.mem.format)
        arr = arr.reshape(self.size[2], self.size[1], self.size[0], -1)
        # Slice it
        slices = []
        for d in reversed(range(3)):
            slices.append(slice(offset[d], offset[d] + size[d]))
        sub_arr = arr[tuple(slices)]
        if pixel_padding is not None:
            padding = np.ones(sub_arr.shape[:3] + (1,), dtype=sub_arr.dtype)
            sub_arr = np.concatenate([sub_arr, pixel_padding * padding], -1)
        return memoryview(np.ascontiguousarray(sub_arr))


def format_from_memoryview(mem, size):
    formatmap = {
        "b": "i1",
        "B": "u1",
        "h": "i2",
        "H": "u2",
        "i": "i4",
        "U": "u4",
        "e": "f2",
        "f": "f4",
    }

    format = str(mem.format)
    format = STRUCT_FORMAT_ALIASES.get(format, format)
    # Process channels
    shape = mem.shape
    collapsed_size = [x for x in size if x > 1]
    if len(shape) == len(collapsed_size) + 1:
        nchannels = shape[-1]
    else:
        assert len(shape) == len(collapsed_size)
        nchannels = 1
    assert 1 <= nchannels <= 4
    if format in ("d", "float64"):
        raise TypeError("GPU's don't support float64 texture formats.")
    elif format not in formatmap:
        raise TypeError(
            f"Cannot convert {format!r} to texture format. Maybe specify format?"
        )
    format = f"{nchannels}x" + formatmap[format]
    return format.lstrip("1x")


# mipmaps: every texture can have a certain number of mipmap levels. Each
# next level is half the size of the previous level. I think we can design our
# API to target level 0 by default and allow a way to upload data to other levels.
# arrays: a d2_array view can be be created from a d2 texture with dept > 1
# cube: a special kind3 of array texture, with six 2D textures.
# cube_array: I suppose you'd have an array of 6xn textures in this case?


class TextureView(Resource):
    """View into a Texture.

    Similar to numpy views, a TextureView is a view into a textures data buffer
    with a (potentially) modified sampling behavior and can specify a
    selection/different view on the texture.

    Passing no more than ``address_mode`` and ``filter`` will create a default
    view on the texture with the given sampling parameters.

    Parameters
    ----------
    texture : Texture
        The texture to view.
    address_mode : str
        How to handle out of bounds access. Use "clamp", "mirror" or "repeat".
        This value can also be set per-channel by providing multiple
        comma-separated values, e.g., "clamp,clamp,repeat".
    filter : str
        Interpolation mode. Possible values are: "nearest" or "linear". This
        value can also be set using three comma-separated values to control mag,
        min and mipmap filters individually, e.g., "linear,linear,nearest".
    format : str
        A format string describing the texture layout. If None, use the same as
        the underlying texture. This must be a pygfx format specifier, e.g.
        "3xf4", but can also be a format specific to the render backend if
        necessary (e.g. from ``wgpu.VertexFormat``).
    view_dim : str
        The dimensionality of the array (1, 2 or 3). If None, use the texture's
        dimension. Or e.g. get a "2d" slice view from a 3d texture, or e.g.
        "cube" or "2d-array".
    aspect : Enum
        The `wgpu.TextureAspect` for this view. Omit or pass None to use the default.
    mip_range : range
        A range object specifying the viewed mip levels.
    layer_range : range
        A range object specifying the array layers to view.

    """

    def __init__(
        self,
        texture,
        *,
        address_mode="clamp",
        filter="nearest",
        format=None,
        view_dim=None,
        aspect=None,
        mip_range=None,
        layer_range=None,
    ):
        super().__init__()
        self._rev = 1
        assert isinstance(texture, Texture)
        self._texture = texture
        # Sampler parameters
        self._address_mode = address_mode
        self._filter = filter
        # Texture view parameters
        self._format = format
        self._view_dim = view_dim
        self._aspect = aspect
        # The ranges
        if mip_range:
            assert isinstance(mip_range, range)
            assert mip_range.step == 1
            self._mip_range = mip_range
        else:
            self._mip_range = None  # None means all mip levels in the texture are used
        if layer_range:
            assert isinstance(layer_range, range)
            assert layer_range.step == 1
            self._layer_range = layer_range
        else:
            self._layer_range = range(texture.size[2])

        self._is_default_view = all(
            x is None for x in [format, view_dim, aspect, mip_range, layer_range]
        )

    @property
    def rev(self):
        # This is not actually increased anywhere, but it's added for consistency
        return self._rev

    @property
    def colorspace(self):
        """Proxy for the texture's colorspace property."""
        return self._texture.colorspace

    @property
    def texture(self):
        """The Texture object holding the data for this texture view."""
        return self._texture

    @property
    def format(self):
        """The texture format."""
        return self._format or self.texture.format

    @property
    def view_dim(self):
        """The dimensionality of this view: "1d", "2d" or "3d"."""
        return self._view_dim or f"{self.texture.dim}d"

    @property
    def mip_range(self):
        """The range of mip levels to view, as a range object.
        The step is always 1.
        """
        return self._mip_range or range(
            self.texture._mip_level_count
        )  # "texture._mip_level_count" may be updated in auto mipmap generation

    @property
    def layer_range(self):
        """The range of array layers to view, as a range object.
        The step is always 1.
        """
        return self._layer_range

    @property
    def address_mode(self):
        """How to sample beyond the edges. Use "clamp",
        "mirror" or "repeat". Default "clamp".
        """
        return self._address_mode

    @property
    def filter(self):
        """Interpolation filter. Use "nearest" or "linear"."""
        return self._filter



from pathlib import Path

import imageio.v3 as iio
from sphinx_gallery.scrapers import figure_rst
from wgpu.gui import WgpuCanvasBase

from ..renderers import Renderer
from .show import Display

# The scraper's default configuration. An example code-block
# may overwrite these values by setting comments of the form
#
#     # sphinx_gallery_pygfx_<name> = <value>
#
# inside the code block. These comments will not be shown in the generated
# gallery example and will be reset after each code block.
default_config = {
    "render": False,  # if True, render an image
    "animate": False,  # if True, render a GIF (TODO)
    "target_name": "renderer",  # the display to use
    # GIF settings
    "duration": 3,  # how many seconds to record
    "loop": 0,  # loop forever
    "lossless": True,  # whether to compress the result
}


def pygfx_scraper(block, block_vars, gallery_conf, **kwargs):
    """Scrape pygfx images and animations

    Parameters
    ----------
    block : tuple
        A tuple containing the (label, content, line_number) of the block.
    block_vars : dict
        Dict of block variables.
    gallery_conf : dict
        Contains the configuration of Sphinx-Gallery
    **kwargs : dict
        Additional keyword arguments to pass to
        :meth:`~matplotlib.figure.Figure.savefig`, e.g. ``format='svg'``.
        The ``format`` kwarg in particular is used to set the file extension
        of the output file (currently only 'png', 'jpg', and 'svg' are
        supported).

    Returns
    -------
    rst : str
        The ReSTructuredText that will be rendered to HTML containing
        the images. This is often produced by :func:`figure_rst`.
    """

    # parse block-level config
    scraper_config = default_config.copy()
    config_prefix = "# sphinx_gallery_pygfx_"
    for line in block[1].split("\n"):
        if not line.startswith(config_prefix):
            continue

        name, value = line[len(config_prefix) :].split(" = ")
        scraper_config[name] = eval(value)

    if not scraper_config["render"] and not scraper_config["animate"]:
        return ""  # nothing to do

    target = block_vars["example_globals"][scraper_config["target_name"]]
    if isinstance(target, Display):
        canvas = target.canvas
    elif isinstance(target, Renderer):
        canvas = target.target
    elif isinstance(target, WgpuCanvasBase):
        canvas = target
    else:
        raise ValueError("`target` must be a Display, Renderer, or Canvas.")

    images = []

    if scraper_config["render"]:
        path_generator = block_vars["image_path_iterator"]
        img_path = next(path_generator)
        iio.imwrite(img_path, canvas.draw())
        images.append(img_path)

    if scraper_config["animate"]:
        frames = []

        # by default videos are rendered at ~ 30 FPS
        n_frames = scraper_config["duration"] * 30
        for _ in range(n_frames):
            frames.append(canvas.draw())

        path_generator = block_vars["image_path_iterator"]
        img_path = Path(next(path_generator)).with_suffix(".webp")
        iio.imwrite(
            img_path,
            frames,
            duration=33,
            loop=scraper_config["loop"],
            lossless=scraper_config["lossless"],
        )
        images.append(img_path)

    return figure_rst(images, gallery_conf["src_dir"])



"""
This module implements text geometry. This is where the text rendering comes
together. Most steps in the text rendering process come from pygfx.utils.text,
though most of the alignment is implemented here.

For details about the text rendering process, see pygfx/utils/text/README.md
"""

import numpy as np

from ._base import Geometry
from ..resources import Buffer
from ..utils import text as textmodule


TEXT_ALIGN_ALTS = {
    "left": "start",
    "top": "start",
    "middle": "center",
    "right": "end",
    "bottom": "end",
}
ANCHOR_X_ALTS = {
    "left": "left",
    "center": "center",
    "middle": "center",
    "right": "right",
}
ANCHOR_Y_ALTS = {
    "top": "top",
    "middle": "middle",
    "center": "middle",
    "baseline": "baseline",
    "bottom": "bottom",
}


WHITESPACE_EXTENTS = {}  # A cache


class TextItem:
    """A formatted piece of text.

    A text item represents a unit piece of text that is formatted in a specific
    way. The TextGeometry converts these into GlyphItem's, and positions these
    so that they together display the intended total text.

    Parameters
    ----------
    text : str
        The text to display.
    font_props : textmodule.FontProps
        Format information for this text item.
    ws_before : str
        Whitespace before the text.
    ws_after : str
        Whitespace after the text.
    allow_break : bool
        If True, allow a linebreak to be placed after this piece of text.

    """

    def __init__(
        self,
        text,
        font_props=None,
        *,
        ws_before="",
        ws_after="",
        allow_break=True,
    ):
        if not isinstance(text, str):
            raise TypeError("TextItem text must be str.")
        if font_props is None:
            font_props = textmodule.FontProps()
        elif not isinstance(font_props, textmodule.FontProps):
            raise TypeError("TextItem font_props must be a FontProps object.")

        self._text = text
        self._font_props = font_props
        self._ws_before = ws_before
        self._ws_after = ws_after
        self._allow_break = bool(allow_break)

    @property
    def text(self):
        """The text for this item."""
        return self._text

    @property
    def font_props(self):
        """The FontProps object to format this text item."""
        return self._font_props

    @property
    def ws_before(self):
        """The whitespace text in front of this item."""
        return self._ws_before

    @property
    def ws_after(self):
        """The whitespace text after this item."""
        return self._ws_after

    @property
    def allow_break(self):
        """Whether or not a line-break is allowed after this item."""
        return self._allow_break


class GlyphItem:
    """A collection of glyphs that represents a unit piece of text.
    Intended for internal use only. In most cases one TextItem results
    in one GlyphItem, but it can be more if multiple fonts are required
    to render the TextItem.
    """

    def __init__(self, positions, indices, meta):
        # Arrays with glyph data
        self.positions = positions
        self.indices = indices
        # Layout data
        self.meta = meta
        self.extent = meta["extent"]
        self.direction = meta["direction"]
        self.ascender = meta["ascender"]
        self.descender = meta["descender"]
        self.allow_break = False
        self.margin_before = 0
        self.margin_after = 0
        # Int offset. Note that this means that a glyph item is bound to a TextGeometry
        self.offset = 0


class TextGeometry(Geometry):
    """Geometry specific for representing text.

    The TextGeometry creates and stores the geometry to render a piece of text.
    It can be provided as plain text or in markdown to support basic formatting.

    Parameters
    ----------
    text : str
        The plain text to render (optional).
    markdown : str
        The text to render, formatted as markdown (optional). See
        ``set_markdown()`` for details on the supported formatting.
    screen_space : bool
        Whether the text is rendered in screen space, in contrast to world
        space.
    font_size : float
        The size of the font, in object coordinates or pixel screen coordinates,
        depending on the value of the ``screen_space`` property. Default 12.
    anchor : str
        The position of the origin of the text. Default "middle-center".
    max_width : float
        The maximum width of the text. Words are wrapped if necessary. A value
        of zero means no wrapping. Default zero.
    line_height : float
        A factor to scale the distance between lines. A value of 1 means the
        "native" font's line distance. Default 1.2.
    text_align : str
        How to align the text. Not implemented.
    family : str, tuple
        The name(s) of the font to prefer. If multiple names are given, they are
        preferred in the given order. Characters that are not supported by any
        of the given fonts are rendered with the default font (from the Noto
        Sans collection).
    direction : str
        The text direction. By default the text direction is determined
        automatically, but is always horizontal. Can be set to 'lrt', 'rtl',
        'ttb' or 'btt'.

    """

    def __init__(
        self,
        text=None,
        *,
        markdown=None,
        screen_space=False,
        font_size=12,
        anchor="middle-center",
        max_width=0,
        line_height=1.2,
        text_align="left",
        family=None,
        direction=None,
    ):
        super().__init__()

        # Init stub buffers
        self.indices = None
        self.positions = None
        self.sizes = None

        # Init props unrelated to layout
        self.screen_space = screen_space
        self._direction = direction

        # Disable layout, so we can initialize first
        self._do_layout = False

        # Check inputs
        inputs = text, markdown
        inputs = [i for i in inputs if i is not None]
        if len(inputs) > 1:
            raise TypeError("Either text or markdown must be given, not both.")

        # Process input
        if text is not None:
            self.set_text(text, family=family)
        elif markdown is not None:
            self.set_markdown(markdown, family=family)
        else:
            self.set_text_items([])

        # Set layout props
        self.font_size = font_size
        self.anchor = anchor
        self.max_width = max_width
        self.line_height = line_height
        self.text_align = text_align

        # Finish layout
        self._do_layout = True
        self.apply_layout()

    @property
    def screen_space(self):
        """Text size unit (screen vs local).

        Returns
        -------
        screen_space : bool
            If False, text size uses the unit of the local frame (e.g. cm).
            Otherwise it is uses the logical screen's units (e.g. px). The
            latter mode is typically used for annotations.

        Notes
        -----
        Regardless of choice, the local object's rotation and scale will still
        transform the text.

        """
        return self._store.screen_space

    @screen_space.setter
    def screen_space(self, value):
        self._store.screen_space = bool(value)

    def set_text_items(self, text_items):
        """Update the text using one or more TextItems.

        .. note::
            This is considered a low level function to provide more control. Use
            ``set_text`` or ``set_markdown`` for more convenience.

        Parameters
        ----------
        text_items : list
            A list of :class:`pygfx.TextItem`s to update the text with.

        Notes
        -----
        If the new text has more glyphs than the current one a new (larger)
        buffer is created. Otherwise, the previous buffers are reused.
        """

        # This function can be considered the core of the text rendering.
        # Everyting comes together here.

        # We cannot have nonzero buffers, so if we have nothing create a single space
        if not text_items:
            text_items = [TextItem(" ")]

        # Convert incoming text items to glyph items
        glyph_items = []
        for item in text_items:
            if not isinstance(item, TextItem):
                raise TypeError("TextGeometry only accepts TextItem objects.")
            first_index = len(glyph_items)

            # Text rendering steps: font selection, shaping, glyph generation
            text_pieces = self._select_font(item.text, item.font_props)
            for text, font in text_pieces:
                glyph_indices, positions, meta = self._shape_text(text, font.filename)
                atlas_indices = self._generate_glyph(glyph_indices, font.filename)
                self._encode_font_props_in_atlas_indices(
                    atlas_indices, item.font_props, font
                )
                glyph_items.append(GlyphItem(positions, atlas_indices, meta))

            # Get whitespace after and before the text
            margin_before = margin_after = 0
            margin_before = self._get_ws_extent(item.ws_before, text_pieces[0][1])
            margin_after = self._get_ws_extent(item.ws_after, text_pieces[-1][1])

            # Set props so these items will be grouped correctly
            first_item, last_item = glyph_items[first_index], glyph_items[-1]
            first_item.allow_break = item.allow_break
            first_item.margin_before = margin_before
            last_item.margin_after = margin_after

        # Layout pre-processing: re-order the items if needed, based on text direction
        i = 0
        while i < len(glyph_items) - 1:
            item = glyph_items[i]
            if item.direction in ("rtl", "btt"):
                i1 = i2 = i
                for j in range(i + 1, len(glyph_items)):
                    if glyph_items[j].direction not in ("rtl", "btt"):
                        break
                    i2 = j
                if i1 != i2:
                    glyph_items[i1 : i2 + 1] = reversed(glyph_items[i1 : i2 + 1])
                i = i2 + 1
            else:
                i += 1

        # We can now store the glyph items
        self._glyph_items = tuple(glyph_items)

        # We set the glyph offsets so we know their place in the total buffer
        glyph_count = 0
        for item in self._glyph_items:
            assert item.indices.dtype == np.uint32
            assert item.positions.dtype == np.float32
            assert item.positions.shape == (item.indices.size, 2)
            item.offset = glyph_count
            glyph_count += item.indices.size

        # Do we need new buffers?
        if self.indices is None or self.indices.nitems < glyph_count:
            self.indices = Buffer(np.zeros((glyph_count,), np.uint32))
            self.positions = Buffer(np.zeros((glyph_count, 2), np.float32))
            self.sizes = Buffer(np.zeros((glyph_count,), np.float32))

        # Copy the glyph arrays into the buffers
        for item in self._glyph_items:
            i1, i2 = item.offset, item.offset + item.indices.shape[0]
            self.indices.data[i1:i2] = item.indices
            self.positions.data[i1:i2] = item.positions

        # Disable the unused space by setting the sizes to zero, leading
        # to degenerate triangles. Leave the indices intact, so that
        # any errors will be detected by the old text shining through.
        self.sizes.data[glyph_count:] = 0

        # Trigger new indices and sizes to be uploaded to the GPU.
        self.sizes.update_range(0, self.indices.nitems)
        self.indices.update_range(0, self.indices.nitems)

        # Finalize the buffers by applying the layout algorithmm.
        self.apply_layout()

    # %%%%% Entrypoint and itemization

    def set_text(self, text, family=None, style=None, weight=None):
        """Update the text.

        Parameters
        ----------
        text : str
            The new text.
        family : str, tuple
            The name(s) of the preferred font(s) to prefer. If multiple names are
            given, they are preferred in the given order. Characters that are
            not supported by any of the given fonts are rendered with the
            default font.
        style : str
            The style of the font (normal, italic, oblique). Default "normal".
        weight : str, int
            The weight of the font. E.g. "normal" or "bold" or a number between
            100 and 900. Default "normal".

        See Also
        --------
        TextGeometry.set_text_items

        """

        if not isinstance(text, str):
            raise TypeError("Text must be a Unicode string.")
        font_props = textmodule.FontProps(family=family, style=style, weight=weight)

        # Split the text in pieces using a tokenizer. We put the
        # whitespace as margin on the text items (whitespace is not rendered)
        items = []
        pending_whitespace = None
        for kind, piece in textmodule.tokenize_text(text):
            if kind == "ws":
                if not items:
                    pending_whitespace = piece
                else:
                    items[-1]._ws_after += piece
            else:
                items.append(TextItem(piece, font_props))
        if items and pending_whitespace:
            items[0]._ws_before += pending_whitespace

        self.set_text_items(items)
        return self

    def set_markdown(self, markdown, family=None):
        """Update the text using markdown formatting.

        The supported subset of markdown is limited to surrounding pieces of
        text with single and double stars for slanted and bold text
        respectively.

        Parameters
        ----------
        markdown : str
            The new text (including markdown).
        family : str, tuple
            The name(s) of the font(s) to prefer. If multiple names are given,
            they are preferred in the given order. Characters that are not
            supported by any of the given fonts are rendered with the default
            font.

        See Also
        --------
        TextGeometry.set_text_items

        """

        if not isinstance(markdown, str):
            raise TypeError("Markdown text must be a Unicode string.")

        # Split text in pieces using a tokenizer
        pieces = list(textmodule.tokenize_markdown(markdown))

        # Put a virtual zero-char space in front and at the end, to make the alg simpler
        pieces.insert(0, ("ws", ""))
        pieces.append(("ws", ""))

        # Prepare font props
        font_props = textmodule.FontProps(family=family)
        pieces_props = [font_props for x in pieces]

        # Now resolve starts to detect bold and italic pieces
        bold_start = slant_start = None
        for i in range(len(pieces)):
            kind, piece = pieces[i]
            if kind == "stars":
                prev_is_wordlike = pieces[i - 1][0] not in ("ws", "punctuation")
                next_is_wordlike = pieces[i + 1][0] not in ("ws", "punctuation")
                if not prev_is_wordlike and next_is_wordlike:
                    # Might be a beginning
                    if piece == "**" and not bold_start:
                        bold_start = i
                    elif piece == "*" and not slant_start:
                        slant_start = i
                elif prev_is_wordlike and not next_is_wordlike:
                    # Might be an end
                    if piece == "**" and bold_start:
                        pieces[bold_start] = "", ""
                        pieces[i] = "", ""
                        for j in range(bold_start + 1, i):
                            pieces_props[j] = pieces_props[j].copy(weight="bold")
                        bold_start = None
                    elif piece == "*" and slant_start:
                        pieces[slant_start] = "", ""
                        pieces[i] = "", ""
                        for j in range(slant_start + 1, i):
                            pieces_props[j] = pieces_props[j].copy(style="slanted")
                        slant_start = None

        # Convert to TextItem objects
        items = []
        pending_whitespace = None
        for i in range(len(pieces)):
            kind, piece = pieces[i]
            if not kind:
                pass
            elif kind == "ws":
                if not items:
                    pending_whitespace = piece
                else:
                    items[-1]._ws_after += piece
            else:
                items.append(TextItem(piece, pieces_props[i]))
        if items and pending_whitespace:
            items[0]._ws_before += pending_whitespace

        self.set_text_items(items)
        return self

    # %%%%% Font selection

    def _select_font(self, text, font_props):
        """The font selection step. Returns (text, font_filename) tuples.
        Can be overloaded for custom behavior.
        """
        return textmodule.select_font(text, font_props)

    # %%%%% Shaping

    def _shape_text(self, text, font_filename):
        """The shaping step. Returns (glyph_indices, positions, meta).
        Can be overloaded for custom behavior.
        """
        return textmodule.shape_text(text, font_filename, self._direction)

    def _get_ws_extent(self, s, font):
        """Get the extent of a piece of whitespace text. Results of small strings are cached."""
        if not s:
            return 0
        elif len(s) <= 8:
            key = (s, self._direction, font.filename)
            try:
                return WHITESPACE_EXTENTS[key]
            except KeyError:
                meta = self._shape_text(s, font.filename)[2]
                extent = meta["extent"]
                WHITESPACE_EXTENTS[key] = extent
                return extent
        else:
            meta = self._shape_text(s, font.filename)[2]
            return meta["extent"]

    # %%%%% Glyph generation

    def _generate_glyph(self, glyph_indices, font_filename):
        """The glyph generation step. Returns an array with atlas indices.
        Can be overloaded for custom behavior.
        """
        return textmodule.generate_glyph(glyph_indices, font_filename)

    def _encode_font_props_in_atlas_indices(self, atlas_indices, font_props, font):
        # We could put font properties in their own buffer(s), but to
        # safe memory, we encode them in the top bits of the atlas
        # indices. This seems like a good place, because these top bits
        # won't be used (2**24 is more than enough glyphs), and the
        # glyph index is a rather "opaque" value to the user anyway.
        # You can think of the new glyph index as the index to the glyph
        # in the atlas, plus props to tweak its appearance.

        # We compare the font_props (i.e. the requested font variant)
        # with the actual font to see what correcion we need to apply.
        slanted_like = "italic", "oblique", "slanted"
        if font_props.style in slanted_like and font.style not in slanted_like:
            atlas_indices += 0x08000000
        weight_offset = font_props.weight - font.weight
        weight_0_15 = int((max(-250, weight_offset) + 250) / 50 + 0.4999)
        atlas_indices += max(0, min(15, weight_0_15)) << 28

    # %%%%% Layout

    def bounding_box(self):
        if self.screen_space:
            # The space occupied by the text is essentially a point
            return np.array([[0, 0, 0], [0, 0, 0]], np.float32)
        else:
            if self._aabb_rev == self.positions.rev:
                return self._aabb
            pos = self.positions.data
            aabb_2d = np.array([pos.min(axis=0), pos.max(axis=0)], np.float32)
            self._aabb[1, 0] += self.font_size  # positions do not include char width
            self._aabb = np.column_stack([aabb_2d, np.zeros((2, 1), np.float32)])
            self._aabb_rev = self.positions.rev
            return self._aabb

    def _apply_layout(self):
        """The layout step. Updates positions and sizes to finalize the geometry.
        Can be overloaded for custom behavior.
        """

        # Prepare

        font_size = self._font_size
        anchor = self._anchor
        positions_array = self.positions.data
        sizes_array = self.sizes.data
        is_horizontal = self._direction is None or self._direction in ("ltr", "rtl")

        left = right = 0
        top = bottom = 0

        # Resolve position and sizes

        extent_offset = 0
        for item in self._glyph_items:
            extent_offset += item.margin_before * font_size
            if is_horizontal:
                positions = item.positions * font_size + np.array([extent_offset, 0])
            else:
                positions = item.positions * font_size + np.array([0, extent_offset])
            i1, i2 = item.offset, item.offset + positions.shape[0]
            positions_array[i1:i2] = positions
            sizes_array[i1:i2] = font_size
            # Prepare for next
            ws_margin = item.margin_after * font_size
            extent_offset += item.extent * font_size + ws_margin
            # Update total extent
            if is_horizontal:
                right = extent_offset
                top = max(top, item.ascender * font_size)
                bottom = min(bottom, item.descender * font_size)
            else:
                bottom = extent_offset
                right = max(right, item.ascender * font_size)
                left = min(left, item.descender * font_size)

        self._aabb = np.array([(left, bottom, 0), (right, top, 0)], np.float32)

        # Anchoring

        if anchor.endswith("left"):
            pos_offset_x = 0
        elif anchor.endswith("center"):
            pos_offset_x = -0.5 * (right - left)
        elif anchor.endswith("right"):
            pos_offset_x = -right

        if anchor.startswith("top"):
            pos_offset_y = -top
        elif anchor.startswith("middle"):
            pos_offset_y = -top + 0.5 * (top - bottom)
        elif anchor.startswith("baseline"):
            pos_offset_y = 0
        elif anchor.startswith("bottom"):
            pos_offset_y = -bottom

        if pos_offset_x or pos_offset_y:
            positions_array += pos_offset_x, pos_offset_y
            self._aabb += pos_offset_x, pos_offset_y, 0

        # Trigger uploads to GPU

        self.sizes.update_range(0, i2)
        self.positions.update_range(0, i2)

    def apply_layout(self):
        """Update the internal contained glyphs.

        To overload this with a custom layout, overload ``_apply_layout()``.
        """

        if self._do_layout:
            self._apply_layout()

    @property
    def font_size(self):
        """The text size.

        For text rendered in screen space (``screen_space`` property is set),
        the size is in logical pixels, and the object's local transform affects
        the final text size.

        For text rendered in world space (``screen_space`` property is *not*
        set), the size is in object coordinates, and the the object's
        world-transform affects the final text size.

        Notes
        -----
        Font size is indicative only. Final glyph size further depends on the
        font family, as glyphs may be smaller (or larger) than the indicative
        size. Final glyph size may further vary based on additional formatting
        applied a particular subsection.

        """
        return self._font_size

    @font_size.setter
    def font_size(self, value):
        self._font_size = float(value)
        self.apply_layout()

    @property
    def max_width(self):
        """The maximum width of the text. Text will wrap if beyond this
        limit. The coordinate system that this applies to is the same
        as for ``font_size``. Set to 0 for no wrap. Default 0.

        TEXT WRAPPING IS NOT YET IMPLEMENTED
        """
        return self._max_width

    @max_width.setter
    def max_width(self, width):
        self._max_width = float(width or 0)
        self.apply_layout()

    @property
    def line_height(self):
        """The relative height of a line of text, used to set the
        distance between lines. Default 1.2. TEXT WRAPPING IS NOT YET IMPLEMENTED
        """
        return self._line_height

    @line_height.setter
    def line_height(self, heigh):
        self._line_height = float(heigh or 1.2)
        self.apply_layout()

    @property
    def text_align(self):
        """Set the alignment of wrapped text. Can be start, end, center, or justify.
        Default "start". TEXT WRAPPING IS NOT YET IMPLEMENTED
        """
        return self._text_align

    @text_align.setter
    def text_align(self, align):
        alignments = "start", "end", "center", "justify"
        if align is None:
            align = "start"
        if not isinstance(align, str):
            raise TypeError("text-align must be a None or str.")
        align = align.lower()
        align = TEXT_ALIGN_ALTS.get(align, align)
        if align not in alignments:
            raise ValueError(f"Align must be one of {alignments}")
        self._text_align = align
        self.apply_layout()

    @property
    def anchor(self):
        """The position of the origin of the text. This is a string
        representing the vertical and horizontal anchors, separated by
        a dash, e.g. "top-left" or "bottom-center".

        * Vertical values: "top", "middle", "baseline", "bottom".
        * Horizontal values: "left", "center", "right".
        """
        return self._anchor

    @anchor.setter
    def anchor(self, anchor):
        # Init
        if anchor is None:
            anchor = "middle-center"
        elif not isinstance(anchor, str):
            raise TypeError("Text anchor must be str.")
        anchor = anchor.lower().strip()
        # Split
        if anchor.count("-") == 1:
            anchory, _, anchorx = anchor.partition("-")
        else:
            anchory = anchorx = ""
            for key, val in ANCHOR_Y_ALTS.items():
                if anchor.startswith(key):
                    anchory = val
                    break
            for key, val in ANCHOR_X_ALTS.items():
                if anchor.endswith(key):
                    anchorx = val
                    break
        # Resolve
        try:
            anchory = ANCHOR_Y_ALTS[anchory]
            anchorx = ANCHOR_X_ALTS[anchorx]
        except KeyError:
            raise ValueError(f"Invalid anchor value '{anchor}'")
        # Apply
        self._anchor = f"{anchory}-{anchorx}"
        self.apply_layout()



"""
The flusher is responsible for rendering the renderers internal image
to the canvas.
"""
import wgpu
import numpy as np


FULL_QUAD_SHADER = """
    struct VertexInput {
        @builtin(vertex_index) index: u32,
    };
    struct Varyings {
        @location(0) texcoord: vec2<f32>,
        @builtin(position) position: vec4<f32>,
    };
    struct FragmentOutput {
        @location(0) color: vec4<f32>,
    };

    BINDINGS_CODE

    @vertex
    fn vs_main(in: VertexInput) -> Varyings {
        var positions = array<vec2<f32>,4>(
            vec2<f32>(0.0, 1.0), vec2<f32>(0.0, 0.0), vec2<f32>(1.0, 1.0), vec2<f32>(1.0, 0.0)
        );
        let pos = positions[in.index];
        var varyings: Varyings;
        varyings.texcoord = vec2<f32>(pos.x, 1.0 - pos.y);
        varyings.position = vec4<f32>(pos * 2.0 - 1.0, 0.0, 1.0);
        return varyings;
    }

    @fragment
    fn fs_main(varyings: Varyings) -> FragmentOutput {
        var out : FragmentOutput;
        let texcoord = varyings.texcoord;  // for textureSample
        let texindex = vec2<i32>(varyings.position.xy);  // for textureLoad

        FRAGMENT_CODE

        return out;
    }
"""


def _create_pipeline(device, binding_layouts, bindings, targets, wgsl):
    shader_module = device.create_shader_module(code=wgsl)

    bind_group_layout = device.create_bind_group_layout(entries=binding_layouts)
    pipeline_layout = device.create_pipeline_layout(
        bind_group_layouts=[bind_group_layout]
    )
    bind_group = device.create_bind_group(layout=bind_group_layout, entries=bindings)

    render_pipeline = device.create_render_pipeline(
        layout=pipeline_layout,
        vertex={
            "module": shader_module,
            "entry_point": "vs_main",
            "buffers": [],
        },
        primitive={
            "topology": wgpu.PrimitiveTopology.triangle_strip,
            "strip_index_format": wgpu.IndexFormat.uint32,
        },
        depth_stencil=None,
        multisample=None,
        fragment={
            "module": shader_module,
            "entry_point": "fs_main",
            "targets": targets,
        },
    )

    return bind_group, render_pipeline


# %%%%%%%%%%


class RenderFlusher:
    """
    Utility to flush (render) the current state of a renderer into a texture.
    """

    def __init__(self, device):
        self._device = device
        self._pipelines = {}

        dtype = [
            ("size", "float32", (2,)),
            ("sigma", "float32"),
            ("support", "int32"),
            ("gamma", "float32"),
            ("_padding", "float32"),
        ]
        self._uniform_data = np.zeros((), dtype=dtype)
        self._uniform_buffer = self._device.create_buffer(
            size=self._uniform_data.nbytes,
            usage=wgpu.BufferUsage.UNIFORM | wgpu.BufferUsage.COPY_DST,
        )

    def render(
        self, src_color_tex, src_depth_tex, dst_color_tex, dst_format, gamma=1.0
    ):
        """Render the (internal) result of the renderer into a texture."""
        # NOTE: cannot actually use src_depth_tex as a sample texture (BindingCollision)
        assert src_depth_tex is None
        assert isinstance(src_color_tex, wgpu.base.GPUTextureView)
        assert isinstance(dst_color_tex, wgpu.base.GPUTextureView)

        # Recreate pipeline? Use ._internal as a true identifier of the texture view
        hash = src_color_tex.size, src_color_tex._internal
        stored_hash = self._pipelines.get(dst_format, ["invalidhash"])[0]
        if hash != stored_hash:
            bind_group, render_pipeline = self._create_pipeline(
                src_color_tex, dst_format
            )
            self._pipelines[dst_format] = hash, bind_group, render_pipeline

        self._update_uniforms(src_color_tex, dst_color_tex, gamma)
        return self._render(dst_color_tex, dst_format)

    def _update_uniforms(self, src_color_tex, dst_color_tex, gamma):
        # Get factor between texture sizes
        factor_x = src_color_tex.size[0] / dst_color_tex.size[0]
        factor_y = src_color_tex.size[1] / dst_color_tex.size[1]
        factor = (factor_x + factor_y) / 2

        if factor == 1:
            # With equal res, we smooth a tiny bit.
            # A bit less crisp, but also less flicker.
            sigma = 0.5
            support = 2
        elif factor > 1:
            # With src a higher res, we will do ssaa.
            # Kernel scales with input res.
            sigma = 0.5 * factor
            support = min(5, int(sigma * 3))
        else:
            # With src a lower res, the output is interpolated.
            # But we also smooth to reduce the blockiness.
            sigma = 0.5
            support = 2

        self._uniform_data["size"] = src_color_tex.size[:2]
        self._uniform_data["sigma"] = sigma
        self._uniform_data["support"] = support
        self._uniform_data["gamma"] = gamma

    def _render(self, dst_color_tex, dst_format):
        device = self._device
        _, bind_group, render_pipeline = self._pipelines[dst_format]

        command_encoder = device.create_command_encoder()

        tmp_buffer = device.create_buffer_with_data(
            data=self._uniform_data,
            usage=wgpu.BufferUsage.COPY_SRC,
        )
        command_encoder.copy_buffer_to_buffer(
            tmp_buffer, 0, self._uniform_buffer, 0, self._uniform_data.nbytes
        )

        render_pass = command_encoder.begin_render_pass(
            color_attachments=[
                {
                    "view": dst_color_tex,
                    "resolve_target": None,
                    "clear_value": (0, 0, 0, 0),
                    "load_op": wgpu.LoadOp.clear,
                    "store_op": wgpu.StoreOp.store,
                }
            ],
            depth_stencil_attachment=None,
        )
        render_pass.set_pipeline(render_pipeline)
        render_pass.set_bind_group(0, bind_group, [], 0, 99)
        render_pass.draw(4, 1)
        render_pass.end()

        return [command_encoder.finish()]

    def _create_pipeline(self, src_texture_view, dst_format):
        device = self._device

        bindings_code = """
            struct Render {
                size: vec2<f32>,
                sigma: f32,
                support: i32,
                gamma: f32,
            };
            @group(0) @binding(0)
            var<uniform> u_render: Render;
            @group(0) @binding(1)
            var r_color: texture_2d<f32>;
        """

        fragment_code = """
            // Get info about the smoothing
            let sigma = u_render.sigma;
            let support = min(5, u_render.support);

            // The reference index is the subpixel index in the source texture that
            // represents the location of this fragment.
            let ref_index = texcoord * u_render.size;

            // For the sampling, we work with integer coords. Also use min/max for the edges.
            let base_index = vec2<i32>(ref_index);
            let min_index = vec2<i32>(0, 0);
            let max_index = vec2<i32>(u_render.size - 1.0);

            // Convolve. Here we apply a Gaussian kernel, the weight is calculated
            // for each pixel individually based on the distance to the ref_index.
            // This means that the textures don't need to align.
            var val: vec4<f32> = vec4<f32>(0.0, 0.0, 0.0, 0.0);
            var weight: f32 = 0.0;
            for (var y:i32 = -support; y <= support; y = y + 1) {
                for (var x:i32 = -support; x <= support; x = x + 1) {
                    let step = vec2<i32>(x, y);
                    let index = clamp(base_index + step, min_index, max_index);
                    let dist = length(ref_index - vec2<f32>(index) - 0.5);
                    let t = dist / sigma;
                    let w = exp(-0.5 * t * t);
                    val = val + textureLoad(r_color, index, 0) * w;
                    weight = weight + w;
                }
            }
            let gamma3 = vec3<f32>(u_render.gamma);
            out.color = vec4<f32>(pow(val.rgb / weight, gamma3), val.a / weight);
        """

        wgsl = FULL_QUAD_SHADER
        wgsl = wgsl.replace("BINDINGS_CODE", bindings_code)
        wgsl = wgsl.replace("FRAGMENT_CODE", fragment_code)

        binding_layouts = [
            {
                "binding": 0,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "buffer": {"type": wgpu.BufferBindingType.uniform},
            },
            {
                "binding": 1,
                "visibility": wgpu.ShaderStage.FRAGMENT,
                "texture": {
                    "sample_type": wgpu.TextureSampleType.unfilterable_float,
                    "view_dimension": wgpu.TextureViewDimension.d2,
                    "multisampled": False,
                },
            },
        ]

        bindings = [
            {
                "binding": 0,
                "resource": {
                    "buffer": self._uniform_buffer,
                    "offset": 0,
                    "size": self._uniform_data.nbytes,
                },
            },
            {"binding": 1, "resource": src_texture_view},
        ]

        targets = [
            {
                "format": dst_format,
                "blend": {
                    "alpha": (
                        wgpu.BlendFactor.src_alpha,
                        wgpu.BlendFactor.one_minus_src_alpha,
                        wgpu.BlendOperation.add,
                    ),
                    "color": (
                        wgpu.BlendFactor.src_alpha,
                        wgpu.BlendFactor.one_minus_src_alpha,
                        wgpu.BlendOperation.add,
                    ),
                },
            },
        ]

        return _create_pipeline(device, binding_layouts, bindings, targets, wgsl)



"""
Utilities to load scenes from files, using trimesh.
"""

import pygfx as gfx


def load_scene(path):
    """Load a scene from a file.

    This function requires the trimesh library. Might not be complete yet.

    Parameters
    ----------
    path : str
        The location where the scene description is stored.

    Returns
    -------
    meshes : list
        A list of loaded meshes.

    """

    import trimesh  # noqa

    scene = trimesh.load(path)
    for node_name in scene.graph.nodes_geometry:
        transform, geometry_name = scene.graph[node_name]
        current = scene.geometry[geometry_name]
        current.apply_transform(transform)

    return [
        gfx.Mesh(
            gfx.geometry_from_trimesh(m),
            gfx.material_from_trimesh(m.visual.material),
        )
        for m in scene.geometry.values()
    ]



# flake8: noqa

"""
Linear Algebra Routines (deprecated)

.. warning::
    This module is deprecated and will be replaced by `pylinalg
    <https://github.com/pygfx/pylinalg>`_ in a future version.

"""

from .utils import *
from .cylindrical import *
from .euler import *
from .matrix3 import *
from .matrix4 import *
from .quaternion import *
from .spherical import *
from .vector3 import *
from .vector4 import *



import random
import weakref
import threading
import enum

import numpy as np

from ..linalg import Vector3, Matrix4, Quaternion
from ..linalg.utils import transform_aabb, aabb_to_sphere
from ..resources import Buffer
from ..utils import array_from_shadertype
from ..utils.trackable import RootTrackable
from ._events import EventTarget


class IdProvider:
    """Object for internal use to manage world object id's."""

    def __init__(self):
        self._ids_in_use = set([0])
        self._map = weakref.WeakValueDictionary()
        self._lock = threading.RLock()

    def claim_id(self, wobject):
        """Used by wobjects to claim an id."""
        # We don't simply count up, but keep a pool of ids. This is
        # because an application *could* create and discard objects at
        # a high rate, so we want to be able to re-use these ids.
        #
        # Some numbers:
        # * 4_294_967_296 (2**32) max number for u32
        # * 2_147_483_647 (2**31 -1) max number for i32.
        # *    16_777_216 max integer that can be stored exactly in f32
        # *     4_000_000 max integer that survives being passed as a varying (in my tests)
        # *     1_048_575 is ~1M is 2**20 seems like a good max scene objects.
        # *    67_108_864 is ~50M is 2**26 seems like a good max vertex count.
        #                 which leaves 64-20-26=18 bits for any other picking info.

        # Max allowed id, inclusive
        id_max = 1_048_575  # 2*20-1

        # The max number of ids. This is a bit less to avoid choking
        # when there are few free id's left.
        max_items = 1_000_000

        with self._lock:
            if len(self._ids_in_use) >= max_items:
                raise RuntimeError("Max number of objects reached.")
            id = 0
            while id in self._ids_in_use:
                id = random.randint(1, id_max)
            self._ids_in_use.add(id)
            self._map[id] = wobject

        return id

    def release_id(self, wobject, id):
        """Release an id associated with a wobject."""
        with self._lock:
            self._ids_in_use.discard(id)
            self._map.pop(id, None)

    def get_object_from_id(self, id):
        """Return the wobject associated with an id, or None."""
        return self._map.get(id)


id_provider = IdProvider()


class RenderMask(enum.IntFlag):
    auto = 0
    opaque = 1
    transparent = 2
    all = 3


class WorldObject(EventTarget, RootTrackable):
    """Base class for objects.

    This class represents objects in the world, i.e., the scene graph.Each
    WorldObject has geometry to define it's data, and material to define its
    appearance. The object itself is only responsible for defining object
    hierarchies (parent / children) and its position and orientation in the
    world.

    Parameters
    ----------
    geometry : Geometry
        The data defining the shape of the object.
    material : Material
        The data defining the appearence of the object.
    visible : bool
        Whether the object is visible.
    render_order : int
        The render order (when applicable for the renderer's blend mode).
    render_mask : str
        Determines the render passes that the object is rendered in. It's
        recommended to let the renderer decide, using "auto".
    position : Vector
        The position of the object in the world. Default (0, 0, 0).

    Notes
    -----
    Use :class:`Group` to collect multiple world objects into a single empty
    world object.

    """

    # The uniform type describes the structured info for this object, which represents
    # every "propery" that a renderer would need to know in order to visualize it.
    # Put larger items first for alignment, also note that host-sharable structs
    # align at power-of-two only, so e.g. vec3 needs padding.
    # todo: rename uniform to info or something?

    uniform_type = dict(
        world_transform="4x4xf4",
        world_transform_inv="4x4xf4",
        id="i4",
    )

    _v = Vector3()
    _m = Matrix4()
    _q = Quaternion()

    def __init__(
        self,
        geometry=None,
        material=None,
        *,
        visible=True,
        render_order=0,
        render_mask="auto",
        position=None,
    ):
        super().__init__()

        self.geometry = geometry
        self.material = material

        # Init visibility and render props
        self.visible = visible
        self.render_order = render_order
        self.render_mask = render_mask

        # Init parent and children
        self._parent_ref = None
        self._children = []

        position = (0, 0, 0) if position is None else position
        self.position = (
            Vector3(*position) if isinstance(position, (tuple, list)) else position
        )
        self.rotation = Quaternion()
        self.scale = Vector3(1, 1, 1)
        self._transform_hash = ()

        self.up = Vector3(0, 1, 0)

        self._matrix = Matrix4()
        self._matrix_auto_update = True
        self._matrix_world = Matrix4()
        self._matrix_world_dirty = True

        # Compose complete uniform type
        self.uniform_type = {}
        for cls in reversed(self.__class__.mro()):
            self.uniform_type.update(getattr(cls, "uniform_type", {}))
        self.uniform_buffer = Buffer(array_from_shadertype(self.uniform_type))

        # Set id
        self._id = id_provider.claim_id(self)
        self.uniform_buffer.data["id"] = self._id

        self.cast_shadow = False
        self.receive_shadow = False

    def __repr__(self):
        return f"<pygfx.{self.__class__.__name__} at {hex(id(self))}>"

    def __del__(self):
        id_provider.release_id(self, self.id)

    @property
    def id(self):
        """An integer id smaller than 2**31 (read-only)."""
        return self._id

    @property
    def tracker(self):
        return self._root_tracker

    @property
    def visible(self):
        """Wheter is object is rendered or not. Default True."""
        return self._store.visible

    @visible.setter
    def visible(self, visible):
        self._store.visible = bool(visible)

    @property
    def render_order(self):
        """This value allows the default rendering order of scene graph
        objects to be controlled. Default 0. See ``Renderer.sort_objects``
        for details.
        """
        return self._store.render_order

    @render_order.setter
    def render_order(self, value):
        self._store.render_order = float(value)

    @property
    def render_mask(self):
        """Indicates in what render passes to render this object:

        * "auto": try to determine the best approach (default).
        * "opaque": only in the opaque render pass.
        * "transparent": only in the transparent render pass(es).
        * "all": render in both opaque and transparent render passses.

        If "auto" (the default), the renderer attempts to determine
        whether all fragments will be either opaque or all transparent,
        and only apply the needed render passes. If this cannot be
        determined, it falls back to "all".

        Some objects may contain both transparent and opaque fragments,
        and should be rendered in all passes - the object's contribution
        to each pass is determined on a per-fragment basis.

        For clarity, rendering objects in all passes even though they
        are fully opaque/transparent yields correct results and is
        generally the "safest" option. The only cost is performance.
        Rendering transparent fragments in the opaque pass makes them
        invisible. Rendering opaque fragments in the transparent pass
        blends them as if they are transparent with an alpha of 1.
        """
        return self._store.render_mask

    @render_mask.setter
    def render_mask(self, value):
        if value is None:
            self._store.render_mask = RenderMask(0)
        elif isinstance(value, int):
            self._store.render_mask = RenderMask(value)
        elif isinstance(value, str):
            try:
                self._store.render_mask = RenderMask._member_map_[value.lower()]
            except KeyError:
                opts = set(RenderMask._member_names_)
                msg = f"WorldObject.render_mask must be one of {opts} not {value!r}"
                raise ValueError(msg) from None
        else:
            raise TypeError(
                f"WorldObject.render_mask must be int or str, not {type(value)}"
            )

    @property
    def geometry(self):
        """The object's geometry, the data that defines (the shape of) this object."""
        return self._store.geometry

    @geometry.setter
    def geometry(self, geometry):
        self._store.geometry = geometry

    @property
    def material(self):
        """Wheter is object is rendered or not. Default True."""
        return self._store.material

    @material.setter
    def material(self, material):
        self._store.material = material

    @property
    def cast_shadow(self):
        """Whether this object casts shadows, i.e. whether it is rendered into
        a shadow map. Default False."""
        return self._cast_shadow  # does not affect any shaders

    @cast_shadow.setter
    def cast_shadow(self, value):
        self._cast_shadow = bool(value)

    @property
    def receive_shadow(self):
        """Whether this object receives shadows. Default False."""
        return self._store.receive_shadow

    @receive_shadow.setter
    def receive_shadow(self, value):
        self._store.receive_shadow = bool(value)

    @property
    def parent(self):
        """Object's parent in the scene graph (read-only).
        An object can have at most one parent.
        """
        return self._parent_ref and self._parent_ref()

    @property
    def children(self):
        """The child objects of this wold object (read-only tuple).
        Use ``.add()`` and ``.remove()`` to change this list.
        """
        return tuple(self._children)

    def add(self, *objects, before=None):
        """Adds object as child of this object. Any number of
        objects may be added. Any current parent on an object passed
        in here will be removed, since an object can have at most one
        parent.
        If ``before`` argument is given (and present in children), then
        the items are inserted before the given element.
        """
        idx = len(self._children)
        if before:
            try:
                idx = self._children.index(before)
            except ValueError:
                pass
        for obj in objects:
            assert isinstance(obj, WorldObject)
            # orphan if needed
            if obj._parent_ref is not None:
                obj._parent_ref().remove(obj)
            # attach to scene graph
            obj._parent_ref = weakref.ref(self)
            self._children.insert(idx, obj)
            idx += 1
            # flag world matrix as dirty
            obj._matrix_world_dirty = True
        return self

    def remove(self, *objects):
        """Removes object as child of this object. Any number of objects may be removed.
        If a given object is not a child, it is ignored.
        """
        for obj in objects:
            try:
                self._children.remove(obj)
                obj._parent_ref = None
            except ValueError:
                pass
        return self

    def clear(self):
        """Removes all children."""
        for child in self._children:
            child._parent_ref = None
        self._children.clear()

    def traverse(self, callback, skip_invisible=False):
        """Executes the callback on this object and all descendants.

        If ``skip_invisible`` is given and True, objects whose
        ``visible`` property is False - and their children - are
        skipped. Note that modifying the scene graph inside the callback
        is discouraged.
        """

        for child in self.iter(skip_invisible=skip_invisible):
            callback(child)

    def iter(self, filter_fn=None, skip_invisible=False):
        """Create a generator that iterates over this objects and its children.
        If ``filter_fn`` is given, only objects for which it returns ``True``
        are included.
        """
        if skip_invisible and not self.visible:
            return

        if filter_fn is None:
            yield self
        elif filter_fn(self):
            yield self

        for child in self._children:
            yield from child.iter(filter_fn, skip_invisible)

    def update_matrix(self):
        p, r, s = self.position, self.rotation, self.scale
        hash = p.x, p.y, p.z, r.x, r.y, r.z, r.w, s.x, s.y, s.z
        if hash != self._transform_hash:
            self._transform_hash = hash
            self._matrix.compose(self.position, self.rotation, self.scale)
            self._matrix_world_dirty = True

    @property
    def matrix(self):
        """The (settable) transformation matrix."""
        return self._matrix

    @matrix.setter
    def matrix(self, matrix):
        self._matrix.copy(matrix)
        self._matrix.decompose(self.position, self.rotation, self.scale)
        self._matrix_world_dirty = True

    @property
    def matrix_world(self):
        """The world matrix (local matrix composed with any parent matrices)."""
        return self._matrix_world

    @property
    def matrix_auto_update(self):
        """Whether or not the matrix auto-updates."""
        return self._matrix_auto_update

    @matrix_auto_update.setter
    def matrix_auto_update(self, value):
        self._matrix_auto_update = bool(value)

    @property
    def matrix_world_dirty(self):
        """Whether or not the matrix needs updating (readonly)."""
        return self._matrix_world_dirty

    def apply_matrix(self, matrix):
        if self._matrix_auto_update:
            self.update_matrix()
        self._matrix.premultiply(matrix)
        self._matrix.decompose(self.position, self.rotation, self.scale)
        self._matrix_world_dirty = True

    def update_matrix_world(
        self, force=False, update_children=True, update_parents=False
    ):
        if update_parents and self.parent:
            self.parent.update_matrix_world(
                force=force, update_children=False, update_parents=True
            )
        if self._matrix_auto_update:
            self.update_matrix()
        if self._matrix_world_dirty or force:
            if self.parent is None:
                self._matrix_world.copy(self._matrix)
            else:
                self._matrix_world.multiply_matrices(
                    self.parent._matrix_world, self._matrix
                )
            self.uniform_buffer.data[
                "world_transform"
            ].flat = self._matrix_world.elements
            tmp_inv_matrix = Matrix4().get_inverse(self._matrix_world)
            self.uniform_buffer.data[
                "world_transform_inv"
            ].flat = tmp_inv_matrix.elements
            self.uniform_buffer.update_range(0, 1)
            self._matrix_world_dirty = False
            for child in self._children:
                child._matrix_world_dirty = True
        if update_children:
            for child in self._children:
                child.update_matrix_world()

    def look_at(self, target: Vector3):
        self.update_matrix_world(update_parents=True, update_children=False)
        self._v.set_from_matrix_position(self._matrix_world)
        self._m.look_at(self._v, target, self.up)
        self.rotation.set_from_rotation_matrix(self._m)
        if self.parent:
            self._m.extract_rotation(self.parent._matrix_world)
            self._q.set_from_rotation_matrix(self._m)
            self.rotation.premultiply(self._q.inverse())

    def get_world_position(self):
        self.update_matrix_world(update_parents=True, update_children=False)
        self._v.set_from_matrix_position(self._matrix_world)
        return self._v.clone()

    def get_world_bounding_box(self):
        """Updates all parent and children world matrices, and returns
        a single world-space axis-aligned bounding box for this object's
        geometry and all of its children (recursively)."""
        self.update_matrix_world(update_parents=True, update_children=True)
        return self._get_world_bounding_box()

    def _get_world_bounding_box(self):
        """Returns a world-space axis-aligned bounding box for this object's
        geometry and all of its children (recursively)."""
        boxes = []
        if self._store.geometry:
            aabb = self._store.geometry.bounding_box()
            aabb_world = transform_aabb(aabb, self._matrix_world.to_ndarray())
            boxes.append(aabb_world)
        if self._children:
            boxes.extend(
                [
                    b
                    for b in (c.get_world_bounding_box() for c in self._children)
                    if b is not None
                ]
            )
        if len(boxes) == 1:
            return boxes[0]
        if boxes:
            boxes = np.array(boxes)
            return np.array([boxes[:, 0].min(axis=0), boxes[:, 1].max(axis=0)])

    def get_world_bounding_sphere(self):
        """Returns a world-space bounding sphere by converting an
        axis-aligned bounding box to a sphere.

        See WorldObject.get_world_bounding_box.
        """
        aabb = self.get_world_bounding_box()
        if aabb is not None:
            return aabb_to_sphere(aabb)

    def _wgpu_get_pick_info(self, pick_value):
        # In most cases the material handles this.
        return self.material._wgpu_get_pick_info(pick_value)



# flake8: noqa

""" Objects to control cameras.

Controllers define how cameras can be interacted with. They are
independent of any GUI toolkits.

.. autosummary::

    Controller
    PanZoomController
    OrbitController
    OrbitOrthoController

"""

from ._base import Controller
from ._orbit import OrbitController, OrbitOrthoController
from ._panzoom import PanZoomController


__all__ = ["Controller", "PanZoomController", "OrbitController", "OrbitOrthoController"]



# flake8: noqa

from . import utils

from .resources import *
from .objects import *
from .geometries import *
from .materials import *
from .cameras import *
from .helpers import *
from .controllers import *

from .renderers import *

from .utils.color import Color
from .utils.load import load_scene
from .utils.show import show, Display
from .utils.viewport import Viewport
from .utils.text import font_manager
from .utils import cm, logger


__version__ = "0.1.11"
version_info = tuple(map(int, __version__.split(".")))

__wgpu_version_range__ = "0.9.0", "0.10.0"


def _test_wgpu_version():
    import wgpu  # noqa

    min_ver, max_ver = __wgpu_version_range__
    min_ver_info = tuple(map(int, min_ver.split(".")))
    max_ver_info = tuple(map(int, max_ver.split(".")))
    detected = f"Detected {wgpu.__version__}, need >={min_ver}, <{max_ver}."
    if wgpu.version_info < min_ver_info:
        logger.error(
            f"Incompatible version of wgpu-py:\n    {detected}\n    To update, use e.g. `pip install -U wgpu`."
        )
    elif wgpu.version_info >= max_ver_info:
        logger.warning(f"Possible incompatible version of wgpu-py:\n    {detected}")


_test_wgpu_version()


def _get_sg_image_scraper():
    import sphinx_gallery.scrapers
    from .utils.gallery_scraper import pygfx_scraper

    # add webp as supported extension
    sphinx_gallery.scrapers._KNOWN_IMG_EXTS += ("webp",)

    return pygfx_scraper


# Elements of this library are derived from three.js, original license
# at time of writing copied here:
# ---
# The MIT License
#
# Copyright © 2010-2022 three.js authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
# ---
# End of original license copy.



import numpy as np

from ..utils.trackable import Trackable
from ..resources import Resource, Buffer, Texture
from ..linalg.utils import aabb_to_sphere


class Geometry(Trackable):
    """Generic container for Geometry data.

    Parameters
    ----------
    kwargs : dict
        A dict of attributes to define on the geometry object. Keys must match
        the naming convention described in the implementation details section of
        the :mod:`Geometries module <pygfx.geometries>`. If they don't they will
        become optional attributes. Values must either be `Resources` or
        ArrayLike.

    Example
    -------

    .. code-block:: py

        g = Geometry(positions=[[1, 2], [2, 4], [3, 5], [4, 1]])
        g.positions.data  # numpy array

    """

    def __init__(self, **kwargs):
        super().__init__()

        self._aabb = None
        self._aabb_rev = None
        self._bsphere = None
        self._bsphere_rev = None

        for name, val in kwargs.items():
            # Get resource object
            if isinstance(val, Resource):
                resource = val
            else:
                if not isinstance(val, np.ndarray):
                    dtype = np.uint32 if name == "indices" else np.float32
                    val = np.asanyarray(val, dtype=dtype)
                if val.dtype == np.float64:
                    raise ValueError(
                        "64-bit float is not supported, use 32-bit floats instead"
                    )
                if name == "grid":
                    dim = val.ndim
                    if dim > 2 and val.shape[-1] <= 4:
                        dim -= 1  # last array dim is probably (a subset of) rgba
                    resource = Texture(val, dim=dim).get_view()
                else:
                    resource = Buffer(val)

            # Checks
            if isinstance(resource, Buffer):
                format = resource.format
                if name == "indices":
                    pass  # No assumptions about shape; they're considered flat anyway
                elif name == "positions":
                    if not format.startswith("3x"):
                        raise ValueError("Expected Nx3 data for positions")
                elif name == "normals":
                    if not format.startswith("3x"):
                        raise ValueError("Expected Nx3 data for normals")
                elif name == "texcoords":
                    if not ("x" not in format or format.startswith(("1x", "2x", "3x"))):
                        raise ValueError("Expected Nx1, Nx2 or Nx3 data for texcoords")
                elif name == "colors":
                    if not format.startswith("4x"):
                        raise ValueError("Expected Nx4 data for colors")
                elif name == "sizes":
                    if not ("x" not in format):
                        raise ValueError("Expected array of scalars for sizes")
                else:
                    pass  # Unknown attribute - no checks

            # Store
            setattr(self, name, resource)

    def __setattr__(self, key, new_value):
        if not key.startswith("_"):
            if isinstance(new_value, Trackable) or key in self._store:
                return setattr(self._store, key, new_value)
        object.__setattr__(self, key, new_value)

    def __getattribute__(self, key):
        if not key.startswith("_"):
            if key in self._store:
                return getattr(self._store, key)
        return object.__getattribute__(self, key)

    def __dir__(self):
        x = object.__dir__(self)
        x.extend(dict.keys(self._store))
        return x

    def bounding_box(self):
        """Compute the axis-aligned bounding box.

        Computes the aabb based on either positions or the shape of the grid
        buffer. If both are present, the bounding box will be computed based on
        the positions buffer.

        Returns
        -------
        aabb : ndarray, [2, 3]
            The axis-aligned bounding box given by the "smallest" (lowest value)
            and "largest" (highest value) corners.

        """
        if hasattr(self, "positions"):
            if self._aabb_rev == self.positions.rev:
                return self._aabb
            pos = self.positions.data
            self._aabb = np.array([pos.min(axis=0), pos.max(axis=0)])
            # If positions contains xy, but not z, assume z=0
            if self._aabb.shape[1] == 2:
                self._aabb = np.column_stack([self._aabb, np.zeros((2, 1), np.float32)])
            self._aabb_rev = self.positions.rev
            return self._aabb

        if hasattr(self, "grid"):
            if self._aabb_rev == self.grid.rev:
                return self._aabb
            # account for multi-channel image data
            if hasattr(self.grid, "texture"):
                # self.grid can be a TextureView instead of a Texture, so in that
                # case, get the shape from the texture attribute
                grid_shape = self.grid.texture.data.shape[: self.grid.texture.dim]
            else:
                grid_shape = self.grid.data.shape[: self.grid.dim]
            # create aabb in index/data space
            aabb = np.array([np.zeros_like(grid_shape), grid_shape], dtype="f8")
            # convert to local image space by aligning
            # center of voxel index (0, 0, 0) with origin (0, 0, 0)
            aabb -= 0.5
            # ensure coordinates are 3D
            # NOTE: important we do this last, we don't want to apply
            # the -0.5 offset to the z-coordinate of 2D images
            if aabb.shape[1] == 2:
                aabb = np.hstack([aabb, [[0], [0]]])
            self._aabb = aabb
            self._aabb_rev = self.grid.rev
            return self._aabb

        raise ValueError(
            "No positions or grid buffer available for bounding volume computation"
        )

    def bounding_sphere(self):
        """Compute a bounding sphere.

        Uses the geometry's axis-aligned bounding box, to estimate a sphere
        which contains the geometry.

        Returns
        -------
        sphere : ndarray, [4]
            A sphere given by it's center and radius. Format: ``(x, y, z, radius)``.

        Notes
        -----
        Since the sphere wraps the geometry's bounding box, it typically won't
        be the minimally binding sphere.

        """
        if self._bsphere is not None and self._bsphere_rev == self._aabb_rev:
            return self._bsphere

        self._bsphere = aabb_to_sphere(self.bounding_box())
        self._bsphere_rev = self._aabb_rev
        return self._bsphere



"""
Containers for buffers and textures.

In pygfx, data is stored in buffers and textures. We collectively call these resources.

.. currentmodule:: pygfx.resources

.. autosummary::
    :toctree: resources/
    :template: ../_templates/custom_layout.rst

    Resource
    Buffer
    Texture
    TextureView

"""

# flake8: noqa

from ._buffer import Resource, Buffer
from ._texture import Texture, TextureView

__all__ = ["Resource", "Buffer", "Texture", "TextureView"]



"""Objects to draw a scene onto a canvas.

The purpose of a renderer is to render (i.e. draw) a scene to a canvas or
texture. It also provides picking, defines the anti-aliasing parameters, and any
post processing effects.

A renderer is directly associated with its target and can only render to that
target. Different renderers can render to the same target though.

It provides a ``.render()`` method that can be called one or more times to
render scenes. This creates a visual representation that is stored internally,
and is finally rendered into its render target (the canvas or texture)::

                              __________
                             | blender  |
    [scenes] -- render() --> |  state   | -- flush() --> [target]
                             |__________|

The internal representation is managed by the blender object. The internal
render textures are typically at a higher resolution to reduce aliasing (SSAA).
The blender has auxilary buffers such as a depth buffer, pick buffer, and
buffers for transparent fragments. Depending on the blend mode, a single render
call may consist of multiple passes (to deal with semi-transparent fragments).

The flush-step resolves the internal representation into the target texture or
canvas, averaging neighbouring fragments for anti-aliasing.

.. currentmodule:: pygfx.renderers

.. autosummary::
    :toctree: renderers/
    :template: ../_templates/custom_layout.rst

    WgpuRenderer
    SvgRenderer
    print_wgpu_report


"""

# flake8: noqa

from ._base import Renderer
from .wgpu import WgpuRenderer, print_wgpu_report
from .svg import SvgRenderer



import wgpu
from ._update import update_buffer
from ._utils import to_vertex_format

from ...objects import PointLight

# todo: idea:
# the shader and a corresponding binding layout for vertex data can also be defined
# on the Shader object. That way, line and point objects can also casts shadows :)


shadow_vertex_shader = """
    struct Matrix4x4 {
        m : mat4x4<f32>
    };
    @group(0) @binding(0) var<uniform> light_view_projection : Matrix4x4;
    @group(1) @binding(0) var<uniform> model_transform : Matrix4x4;
    @vertex
    fn vs_main(@location(0) position: vec3<f32>) -> @builtin(position) vec4<f32> {
        return light_view_projection.m * model_transform.m * vec4<f32>(position, 1.0);
    }
"""

binding_layout = [
    {
        "binding": 0,
        "visibility": wgpu.ShaderStage.VERTEX,
        "buffer": {"type": wgpu.BufferBindingType.uniform},
    }
]


class ShadowUtil:
    def __init__(self, device):
        self.device = device
        self.pipelines = {}
        self.bind_group_layout = self.device.create_bind_group_layout(
            entries=binding_layout
        )

    def get_shadow_pipeline(self, wobject):
        # shadow pipeline only depends on the object's geometry info (Vertex Buffer)
        # TODO: now it only depends on the positions, but for instanced meshes, it also needs instance buffer
        #  Maybe for future skinned meshes , morph buffer is also needed
        positions = wobject.geometry.positions

        array_stride = positions.nbytes // positions.nitems
        vertex_format = to_vertex_format(positions.format)

        hash = (array_stride, vertex_format)

        if hash not in self.pipelines:
            self.pipelines[hash] = self._create_shadow_pipeline(
                array_stride, vertex_format
            )

        return self.pipelines[hash]

    def _create_shadow_pipeline(self, array_stride, vertex_format):
        vertex_buffer_descriptor = [
            {
                "array_stride": array_stride,
                "step_mode": wgpu.VertexStepMode.vertex,  # vertex
                "attributes": [
                    {
                        "format": vertex_format,
                        "offset": 0,
                        "shader_location": 0,
                    }
                ],
            }
        ]

        program = self.device.create_shader_module(code=shadow_vertex_shader)
        pipeline = self.device.create_render_pipeline(
            layout=self.device.create_pipeline_layout(
                bind_group_layouts=[self.bind_group_layout, self.bind_group_layout]
            ),
            vertex={
                "module": program,
                "entry_point": "vs_main",
                "buffers": vertex_buffer_descriptor,
            },
            primitive={
                "topology": wgpu.PrimitiveTopology.triangle_list,
                "cull_mode": "front",
            },
            depth_stencil={
                "format": wgpu.TextureFormat.depth32float,
                "depth_write_enabled": True,
                "depth_compare": wgpu.CompareFunction.less,
                "stencil_read_mask": 0,
                "stencil_write_mask": 0,
            },
        )

        return pipeline

    def render_shadow_maps(self, lights, wobjects, command_encoder):
        """Render the shadow map for the given lights and wobjects."""

        for light in lights:
            if light.cast_shadow:
                if isinstance(light, PointLight):
                    for buffer in light.shadow._gfx_matrix_buffer:
                        update_buffer(self.device, buffer)
                else:
                    update_buffer(self.device, light.shadow._gfx_matrix_buffer)

                if not isinstance(light.shadow._wgpu_tex_view, list):
                    shadow_maps = [light.shadow._wgpu_tex_view]
                else:
                    shadow_maps = light.shadow._wgpu_tex_view

                for i, shadow_map in enumerate(shadow_maps):
                    shadow_pass = command_encoder.begin_render_pass(
                        color_attachments=[],
                        depth_stencil_attachment={
                            "view": shadow_map,
                            "depth_read_only": False,
                            "depth_clear_value": 1.0,
                            "depth_load_op": wgpu.LoadOp.clear,
                            "depth_store_op": wgpu.StoreOp.store,
                            "stencil_read_only": True,
                            "stencil_load_op": wgpu.LoadOp.clear,
                            "stencil_store_op": wgpu.StoreOp.discard,
                        },
                    )

                    if not hasattr(light.shadow, f"__shadow_bind_group_{i}"):
                        buffer = (
                            light.shadow._gfx_matrix_buffer[i]._wgpu_buffer[1]
                            if isinstance(light, PointLight)
                            else light.shadow._gfx_matrix_buffer._wgpu_buffer[1]
                        )

                        setattr(
                            light.shadow,
                            f"__shadow_bind_group_{i}",
                            self.device.create_bind_group(
                                layout=self.bind_group_layout,
                                entries=[
                                    {
                                        "binding": 0,
                                        "resource": {
                                            "buffer": buffer,
                                            "offset": 0,
                                            "size": 64,
                                        },
                                    }
                                ],
                            ),
                        )

                    shadow_pass.set_bind_group(
                        0, getattr(light.shadow, f"__shadow_bind_group_{i}"), [], 0, 99
                    )

                    for wobject in wobjects:
                        if wobject.cast_shadow and wobject.geometry is not None:
                            shadow_pipeline = self.get_shadow_pipeline(wobject)

                            if shadow_pipeline is not None:
                                shadow_pass.set_pipeline(shadow_pipeline)

                                position_buffer = wobject.geometry.positions

                                shadow_pass.set_vertex_buffer(
                                    0,
                                    position_buffer._wgpu_buffer[1],
                                    position_buffer.vertex_byte_range[0],
                                    position_buffer.vertex_byte_range[1],
                                )

                                if not hasattr(wobject, f"__shadow_bind_group"):
                                    # Note that here we assume that the wobject's transform matrix
                                    # is the first item in its uniform buffer. This is a likely
                                    # assumption because items are sorted by size.
                                    bg = self.device.create_bind_group(
                                        layout=self.bind_group_layout,
                                        entries=[
                                            {
                                                "binding": 0,
                                                "resource": {
                                                    "buffer": wobject.uniform_buffer._wgpu_buffer[
                                                        1
                                                    ],
                                                    "offset": 0,
                                                    "size": 64,
                                                },
                                            }
                                        ],
                                    )

                                    setattr(wobject, "__shadow_bind_group", bg)

                                shadow_pass.set_bind_group(
                                    1,
                                    getattr(wobject, "__shadow_bind_group"),
                                    [],
                                    0,
                                    99,
                                )

                                ibuffer = wobject.geometry.indices

                                n = wobject.geometry.indices.data.size
                                n_instance = 1  # not support instanced meshes yet

                                if ibuffer is not None:
                                    index_format = to_vertex_format(ibuffer.format)
                                    index_format = index_format.split("x")[0].replace(
                                        "s", "u"
                                    )
                                    shadow_pass.set_index_buffer(
                                        ibuffer._wgpu_buffer[1], index_format
                                    )
                                    shadow_pass.draw_indexed(n, n_instance)
                                else:
                                    shadow_pass.draw(n, n_instance)

                    shadow_pass.end()



import sys
import subprocess


def _determine_can_use_wgpu_lib():
    # For some reason, since wgpu-native 5c304b5ea1b933574edb52d5de2d49ea04a053db
    # the process' exit code is not zero, so we test more pragmatically.
    code = "import wgpu.utils; wgpu.utils.get_default_device(); print('ok')"
    result = subprocess.run(
        [
            sys.executable,
            "-c",
            code,
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    print("_determine_can_use_wgpu_lib() status code:", result.returncode)
    return (
        result.stdout.strip().endswith("ok")
        and "traceback" not in result.stderr.lower()
    )


can_use_wgpu_lib = _determine_can_use_wgpu_lib()



import wgpu  # only for flags/enums

from . import register_wgpu_render_function, WorldObjectShader, Binding, RenderMask
from ...utils import array_from_shadertype
from ...resources import Buffer
from ...objects import Line
from ...materials import (
    LineMaterial,
    LineThinMaterial,
    LineThinSegmentMaterial,
    LineSegmentMaterial,
    LineArrowMaterial,
)


# ## Notes
#
# Rendering lines on the GPU is hard! The approach used here uses
# VertexId and storage buffers instead of vertex buffers. That way we
# can create multiple vertices for each point on the line. These
# vertices are positioned such that rendering with triangle_strip
# results in a tick line. More info below.
#
# An alternative (used by e.g. ThreeJS) is to draw the geometry of a
# line segment many times using instancing. Note that some Intel drivers
# limit the number of instances to about 5-10k, which is not much considering
# that lines consisting of millions of points can normally run realtime.
# https://wwwtyro.net/2019/11/18/instanced-lines.html
#
# Another alternative is a geometry shader, but wgpu does not have that.
# But we have compute shaders, which can be used to prepare the triangle-based
# geometry and store it in a buffer, which can be used as a vertex buffer in
# the render pass. The downside of this approach is the memory consumption.
# But a hybrid approach may be viable: the current approach using VertexId,
# but preparing some data in a buffer.

# todo: we can learn about dashing, unfolding and more at http://jcgt.org/published/0002/02/08/paper.pdf


renderer_uniform_type = dict(last_i="i4")


@register_wgpu_render_function(Line, LineMaterial)
class LineShader(WorldObjectShader):
    type = "render"

    def __init__(self, wobject):
        super().__init__(wobject)
        self["line_type"] = "line"
        self["aa"] = wobject.material.aa

    def get_bindings(self, wobject, shared):
        material = wobject.material
        geometry = wobject.geometry

        self["vertex_color_channels"] = 0

        positions1 = geometry.positions

        # With vertex buffers, if a shader input is vec4, and the vbo has
        # Nx2, the z and w element will be zero. This works, because for
        # vertex buffers we provide additional information about the
        # striding of the data.
        # With storage buffers (aka SSBO) we just have some bytes that we
        # read from/write to in the shader. This is more free, but it means
        # that the data in the buffer must match with what the shader
        # expects. In addition to that, there's this thing with vec3's which
        # are padded to 16 bytes. So we either have to require our users
        # to provide Nx4 data, or read them as an array of f32.
        # Anyway, extra check here to make sure the data matches!
        if positions1.data.shape[1] != 3:
            raise ValueError(
                "For rendering (thick) lines, the geometry.positions must be Nx3."
            )

        uniform_buffer = Buffer(array_from_shadertype(renderer_uniform_type))
        uniform_buffer.data["last_i"] = positions1.nitems - 1

        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("u_renderer", "buffer/uniform", uniform_buffer),
            Binding("s_positions", rbuffer, positions1, "VERTEX"),
        ]

        # Per-vertex color, colormap, or a plane color?
        self["color_mode"] = "uniform"
        if material.vertex_colors:
            self["color_mode"] = "vertex"
            self["vertex_color_channels"] = nchannels = geometry.colors.data.shape[1]
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif material.map is not None:
            self["color_mode"] = "map"
            bindings.extend(
                self.define_vertex_colormap(material.map, geometry.texcoords)
            )

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.triangle_strip,
            "cull_mode": wgpu.CullMode.none,
        }

    def _get_n(self, positions):
        return positions.nitems * 5

    def get_render_info(self, wobject, shared):
        material = wobject.material
        # Determine how many vertices are needed
        n = self._get_n(wobject.geometry.positions)
        # Determine in what render passes this objects must be rendered
        render_mask = wobject.render_mask
        if not render_mask:
            render_mask = RenderMask.all
            if material.is_transparent:
                render_mask = RenderMask.transparent
            elif self["color_mode"] == "uniform":
                if material.color_is_transparent:
                    render_mask = RenderMask.transparent
                elif material.aa:
                    render_mask = RenderMask.all
                else:
                    render_mask = RenderMask.opaque
            elif self["color_mode"] == "vertex":
                if self["vertex_color_channels"] in (2, 4):
                    render_mask = RenderMask.all
                elif material.aa:
                    render_mask = RenderMask.all
                else:
                    render_mask = RenderMask.opaque
            elif self["color_mode"] == "map":
                if self["colormap_nchannels"] in (2, 4):
                    render_mask = RenderMask.all
                elif material.aa:
                    render_mask = RenderMask.all
                else:
                    render_mask = RenderMask.opaque
        return {
            "indices": (n, 1),
            "render_mask": render_mask,
        }

    def get_code(self):
        return (
            self.code_definitions()
            + self.code_more_definitions()
            + self.code_common()
            + self.code_helpers()
            + self.code_vertex()
            + self.code_fragment()
        )

    def code_more_definitions(self):
        return """
        struct VertexInput {
            @builtin(vertex_index) index : u32,
        };

        struct VertexFuncOutput {
            i: i32,
            pos: vec4<f32>,
            thickness_p: f32,
            vec_from_node_p: vec2<f32>,
        };

        struct FragmentOutput {
            @location(0) color: vec4<f32>,
            @location(1) pick: vec4<i32>,
            @builtin(frag_depth) depth : f32,
        };
        """

    def code_helpers(self):
        return """

        fn get_point_ndc(index:i32) -> vec4<f32> {
            let raw_pos = load_s_positions(index);
            let world_pos = u_wobject.world_transform * vec4<f32>(raw_pos.xyz, 1.0);
            return u_stdinfo.projection_transform * u_stdinfo.cam_transform * world_pos;
        }
        """

    def code_vertex(self):
        core = self.code_vertex_core()

        return (
            core
            + """
        @vertex
        fn vs_main(in: VertexInput) -> Varyings {

            let index = i32(in.index);
            let screen_factor = u_stdinfo.logical_size.xy / 2.0;
            let l2p:f32 = u_stdinfo.physical_size.x / u_stdinfo.logical_size.x;
            let extra_thick = {{ '0.5' if aa else '0.0' }} / l2p;
            let half_thickness:f32 = u_material.thickness * 0.5 + extra_thick;  // logical pixels

            let result: VertexFuncOutput = get_vertex_result(index, screen_factor, half_thickness, l2p);
            let i0 = result.i;

            var varyings: Varyings;
            varyings.position = vec4<f32>(result.pos);
            varyings.world_pos = vec3<f32>(ndc_to_world_pos(result.pos));
            varyings.thickness_p = f32(result.thickness_p);
            varyings.vec_from_node_p = vec2<f32>(result.vec_from_node_p);

            // Picking
            // Note: in theory, we can store ints up to 16_777_216 in f32,
            // but in practice, its about 4_000_000 for f32 varyings (in my tests).
            // We use a real u32 to not lose presision, see frag shader for details.
            varyings.pick_idx = u32(result.i);
            varyings.pick_zigzag = f32(select(0.0, 1.0, result.i % 2 == 0));

            // Per-vertex colors
            $$ if vertex_color_channels == 1
            let cvalue = load_s_colors(i0);
            varyings.color = vec4<f32>(cvalue, cvalue, cvalue, 1.0);
            $$ elif vertex_color_channels == 2
            let cvalue = load_s_colors(i0);
            varyings.color = vec4<f32>(cvalue.r, cvalue.r, cvalue.r, cvalue.g);
            $$ elif vertex_color_channels == 3
            varyings.color = vec4<f32>(load_s_colors(i0), 1.0);
            $$ elif vertex_color_channels == 4
            varyings.color = vec4<f32>(load_s_colors(i0));
            $$ endif

            // Set texture coords
            $$ if colormap_dim == '1d'
            varyings.texcoord = f32(load_s_texcoords(i0));
            $$ elif colormap_dim == '2d'
            varyings.texcoord = vec2<f32>(load_s_texcoords(i0));
            $$ elif colormap_dim == '3d'
            varyings.texcoord = vec3<f32>(load_s_texcoords(i0));
            $$ endif

            return varyings;
        }
        """
        )

    def code_vertex_core(self):
        return """
        fn get_vertex_result(
            index:i32, screen_factor:vec2<f32>, half_thickness:f32, l2p:f32
        ) -> VertexFuncOutput {
            // This vertex shader uses VertexId and storage buffers instead of
            // vertex buffers. It creates 5 vertices for each point on the line.
            // You could do with 4 to get bevel joins, but round joins are so
            // much nicer. The extra vertex is used to cover more fragments at
            // the joins (and caps). In the fragment shader we discard fragments
            // that are "out of range", based on a varying that represents the
            // vector from the node to the vertex.
            //
            // Basic algorithm and definitions:
            //
            // - We read the positions of three nodes, the current, previous, and next.
            // - These are converted to logical pixel screen space.
            // - We define four normal vectors (na, nb, nc, nd) which represent the
            //   vertices. Two for the previous segment, two for the next. One extra
            //   normal/vertex is defined at the join.
            // - These calculations are done for each vertex (yeah, bit of a waste),
            //   we select just one as output.
            //
            //            /  o     node 3
            //           /  /  /
            //          d  /  /
            //   - - - b  /  /     corners rectangles a, b, c, d
            //   o-------o  /      the vertex e is extra
            //   - - - - a c
            //                node 2
            //  node 1
            //
            // Note that at the inside of a join, the normals (b and d above)
            // move past each-other, causing the rectangles of both segments to
            // overlap. We could prevent this, but that would screw up
            // v_vec_from_node_p. Furthermore, consider a thick line with dense
            // nodes jumping all over the place, causing a lot of overlap anyway.
            // In summary, this viz relies on depth testing with "less" (or
            // semi-tranparent lines would break).
            //
            // Possible improvents:
            //
            // - can we do dashes/stipling?
            // - also implement bevel and miter joins
            // - also implement different caps
            // - we can prepare the nodes' screen coordinates in a compute shader.

            let i = index / 5;

            // Sample the current node and it's two neighbours, and convert to NDC
            // Note that if we sample out of bounds, this affects the shader in mysterious ways (21-12-2021).
            let npos1 = get_point_ndc(max(0, i - 1));
            let npos2 = get_point_ndc(i);
            let npos3 = get_point_ndc(min(u_renderer.last_i, i + 1));

            // Convert to logical screen coordinates, because that's where the lines work
            let ppos1 = (npos1.xy / npos1.w + 1.0) * screen_factor;
            let ppos2 = (npos2.xy / npos2.w + 1.0) * screen_factor;
            let ppos3 = (npos3.xy / npos3.w + 1.0) * screen_factor;

            // Get vectors normal to the line segments
            var v1: vec2<f32> = ppos2.xy - ppos1.xy;
            var v2: vec2<f32> = ppos3.xy - ppos2.xy;
            var na: vec2<f32>;
            var nb: vec2<f32>;
            var nc: vec2<f32>;
            var nd: vec2<f32>;
            var ne: vec2<f32>;

            if (i == 0) {
                // This is the first point on the line: create a cap.
                v1 = v2;
                nc = normalize(vec2<f32>(v2.y, -v2.x));
                nd = -nc;
                na = nd;
                nb = nd - normalize(v2);
                ne = nc - normalize(v2);
            } else if (i == u_renderer.last_i) {
                // This is the last point on the line: create a cap.
                v2 = v1;
                na = normalize(vec2<f32>(v1.y, -v1.x));
                nb = -na;
                ne = nb + normalize(v1);
                nc = na + normalize(v1);
                nd = na;
            } else {
                // Create a join
                na = normalize(vec2<f32>(v1.y, -v1.x));
                nb = -na;
                nc = normalize(vec2<f32>(v2.y, -v2.x));
                nd = -nc;

                // Determine the angle between two of the normals. If this angle is smaller
                // than zero, the inside of the join is at nb/nd, otherwise it is at na/nc.
                let angle = -atan2( na.x * nc.y - na.y * nc.x, na.x * nc.x + na.y * nc.y );

                // Determine the direction of ne
                let vec_dir = select(na + nc, nb + nd, angle >= 0.0);

                // From the angle we can also determine how long the ne vector should be.
                // We express it in a vector magnifier, and limit it to a factor 2,
                // since when the angle is ~pi, the intersection is near infinity.
                // For a bevel join we can omit ne (or set vec_mag to 1.0).
                // For a miter join we'd need an extra vertex to smoothly transition
                // from a miter to a bevel when the angle is too small.
                let vec_mag = 1.0 / max(0.25, cos(0.5 * angle));

                ne = normalize(vec_dir) * vec_mag;
            }

            // Select the correct vector, note that all except ne are unit.
            var vectors = array<vec2<f32>,5>(na, nb, ne, nc, nd);
            let the_vec = vectors[index % 5] * half_thickness;
            let the_pos = ppos2 + the_vec;

            var out : VertexFuncOutput;
            out.i = i;
            out.pos = vec4<f32>((the_pos / screen_factor - 1.0) * npos2.w, npos2.zw);
            out.thickness_p = half_thickness * 2.0 * l2p;
            out.vec_from_node_p = the_vec * l2p;
            return out;
        }
        """

    def code_fragment(self):
        return """
        @fragment
        fn fs_main(varyings: Varyings) -> FragmentOutput {

            // Discard fragments outside of the radius. This is what makes round
            // joins and caps. If we ever want bevel or miter joins, we should
            // change the vertex positions a bit, and drop these lines below.
            let dist_to_node_p = length(varyings.vec_from_node_p);
            if (dist_to_node_p > varyings.thickness_p * 0.5) {
                discard;
            }

            // Prep
            var alpha: f32 = 1.0;

            // Anti-aliasing. Note that because of the discarding above, we cannot use MSAA.
            // By default, the renderer uses SSAA (super-sampling), but if we apply AA for the edges
            // here this will help the end result. Because this produces semitransparent fragments,
            // it relies on a good blend method, and the object gets drawn twice.
            $$ if aa
                let aa_width = 1.0;
                alpha = ((0.5 * varyings.thickness_p) - abs(dist_to_node_p)) / aa_width;
                alpha = clamp(alpha, 0.0, 1.0);
            $$ endif

            // Set color
            $$ if color_mode == 'vertex'
                let color = varyings.color;
            $$ elif color_mode == 'map'
                let color = sample_colormap(varyings.texcoord);
            $$ else
                let color = u_material.color;
            $$ endif

            let physical_color = srgb2physical(color.rgb);
            let opacity = min(1.0, color.a) * alpha * u_material.opacity;
            let out_color = vec4<f32>(physical_color, opacity);

            // Wrap up
            apply_clipping_planes(varyings.world_pos);
            var out = get_fragment_output(varyings.position.z, out_color);

            // Set picking info.
            $$ if write_pick
            // The wobject-id must be 20 bits. In total it must not exceed 64 bits.
            // The pick_idx is int-truncated, so going from a to b, it still has the value of a
            // even right up to b. The pick_zigzag alternates between 0 (even indices) and 1 (odd indices).
            // Here we decode that. The result is that we can support vertex indices of ~32 bits if we want.
            let is_even = varyings.pick_idx % 2u == 0u;
            var coord = select(varyings.pick_zigzag, 1.0 - varyings.pick_zigzag, is_even);
            coord = select(coord, coord - 1.0, coord > 0.5);
            let idx = varyings.pick_idx + select(0u, 1u, coord < 0.0);
            out.pick = (
                pick_pack(u32(u_wobject.id), 20) +
                pick_pack(u32(idx), 26) +
                pick_pack(u32(coord * 100000.0 + 100000.0), 18)
            );
            $$ endif

            // The outer edges with lower alpha for aa are pushed a bit back to avoid artifacts.
            // This is only necessary for blend method "ordered1"
            //out.depth = varyings.position.z + 0.0001 * (0.8 - min(0.8, alpha));

            return out;
        }
        """


@register_wgpu_render_function(Line, LineSegmentMaterial)
class LineSegmentShader(LineShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["line_type"] = "segment"

    def _get_n(self, positions):
        return (positions.nitems // 2) * 2 * 5

    def code_vertex_core(self):
        return """
        fn get_vertex_result(
            index:i32, screen_factor:vec2<f32>, half_thickness:f32, l2p:f32
        ) -> VertexFuncOutput {
            // Similar to the regular line shader, except we only draw segments,
            // using 5 vertices per node. Four for the segments, and 1 to create
            // a degenerate triangle for the space in between. So we only draw
            // caps, no joins.

            let i = index / 5;

            // Sample the current node and either of its neighbours
            let i3 = i + 1 - (i % 2) * 2;  // (i + 1) if i is even else (i - 1)
            let npos2 = get_point_ndc(i);
            let npos3 = get_point_ndc(i3);
            // Convert to logical screen coordinates, because that's were the lines work
            let ppos2 = (npos2.xy / npos2.w + 1.0) * screen_factor;
            let ppos3 = (npos3.xy / npos3.w + 1.0) * screen_factor;

            // Get vectors normal to the line segments
            var na: vec2<f32>;
            var nb: vec2<f32>;
            var nc: vec2<f32>;
            var nd: vec2<f32>;

            // Get vectors normal to the line segments
            if ((i % 2) == 0) {
                // A left-cap
                let v = normalize(ppos3.xy - ppos2.xy);
                nc = vec2<f32>(v.y, -v.x);
                nd = -nc;
                na = nc - v;
                nb = nd - v;
            } else {
                // A right cap
                let v = normalize(ppos2.xy - ppos3.xy);
                na = vec2<f32>(v.y, -v.x);
                nb = -na;
                nc = na + v;
                nd = nb + v;
            }

            // Select the correct vector
            // Note the replicated vertices to create degenerate triangles
            var vectors = array<vec2<f32>,10>(na, na, nb, nc, nd, na, nb, nc, nd, nd);
            let the_vec = vectors[index % 10] * half_thickness;
            let the_pos = ppos2 + the_vec;

            var out : VertexFuncOutput;
            out.i = i;
            out.pos = vec4<f32>((the_pos / screen_factor - 1.0) * npos2.w, npos2.zw);
            out.thickness_p = half_thickness * 2.0 * l2p;
            out.vec_from_node_p = the_vec * l2p;
            return out;
        }
        """


@register_wgpu_render_function(Line, LineArrowMaterial)
class LineArrowShader(LineShader):
    def __init__(self, wobject):
        super().__init__(wobject)
        self["line_type"] = "arrow"

    def _get_n(self, positions):
        return (positions.nitems // 2) * 2 * 4

    def code_vertex_core(self):
        return """
        fn get_vertex_result(
            index:i32, screen_factor:vec2<f32>, half_thickness:f32, l2p:f32
        ) -> VertexFuncOutput {
            // Similar to the normal vertex shader, except we only draw segments,
            // using 3 vertices per node: 6 per segment. 4 for the arrow, and 2
            // to create a degenerate triangle for the space in between. So we
            // only draw caps, no joins.

            let i = index / 3;

            // Sample the current node and either of its neighbours
            let i3 = i + 1 - (i % 2) * 2;  // (i + 1) if i is even else (i - 1)
            let npos2 = get_point_ndc(i);
            let npos3 = get_point_ndc(i3);
            // Convert to logical screen coordinates, because that's were the lines work
            let ppos2 = (npos2.xy / npos2.w + 1.0) * screen_factor;
            let ppos3 = (npos3.xy / npos3.w + 1.0) * screen_factor;

            // Get vectors normal to the line segments
            var na: vec2<f32>;
            var nb: vec2<f32>;

            // Get vectors normal to the line segments
            if ((i % 2) == 0) {
                // A left-cap
                let v = ppos3.xy - ppos2.xy;
                na = normalize(vec2<f32>(v.y, -v.x)) * half_thickness;
                nb = v;
            } else {
                // A right cap
                let v = ppos2.xy - ppos3.xy;
                na = -0.75 * v;
                nb = normalize(vec2<f32>(-v.y, v.x)) * half_thickness - v;
            }

            // Select the correct vector
            // Note the replicated vertices to create degenerate triangles
            var vectors = array<vec2<f32>,6>(na, na, nb, na, nb, nb);
            let the_vec = vectors[index % 6];
            let the_pos = ppos2 + the_vec;

            var out : VertexFuncOutput;
            out.i = i;
            out.pos = vec4<f32>((the_pos / screen_factor - 1.0) * npos2.w, npos2.zw);
            out.thickness_p = half_thickness * 2.0 * l2p;
            out.vec_from_node_p = vec2<f32>(0.0, 0.0);
            return out;
        }
        """


@register_wgpu_render_function(Line, LineThinMaterial)
class ThinLineShader(WorldObjectShader):
    type = "render"

    def get_bindings(self, wobject, shared):
        material = wobject.material
        geometry = wobject.geometry

        rbuffer = "buffer/read_only_storage"
        bindings = [
            Binding("u_stdinfo", "buffer/uniform", shared.uniform_buffer),
            Binding("u_wobject", "buffer/uniform", wobject.uniform_buffer),
            Binding("u_material", "buffer/uniform", material.uniform_buffer),
            Binding("s_positions", rbuffer, geometry.positions, "VERTEX"),
        ]

        # Per-vertex color, colormap, or a plane color?
        self["vertex_color_channels"] = 0
        self["color_mode"] = "uniform"
        if material.vertex_colors:
            self["color_mode"] = "vertex"
            self["vertex_color_channels"] = nchannels = geometry.colors.data.shape[1]
            if nchannels not in (1, 2, 3, 4):
                raise ValueError(f"Geometry.colors needs 1-4 columns, not {nchannels}")
            bindings.append(Binding("s_colors", rbuffer, geometry.colors, "VERTEX"))
        elif material.map is not None:
            self["color_mode"] = "map"
            bindings.extend(
                self.define_vertex_colormap(material.map, geometry.texcoords)
            )

        bindings = {i: b for i, b in enumerate(bindings)}
        self.define_bindings(0, bindings)

        return {
            0: bindings,
        }

    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.line_strip,
            "cull_mode": wgpu.CullMode.none,
        }

    def get_render_info(self, wobject, shared):
        render_mask = wobject.render_mask
        if not render_mask:
            render_mask = RenderMask.all
            if wobject.material.is_transparent:
                render_mask = RenderMask.transparent
            if self["color_mode"] == "uniform":
                if wobject.material.color_is_transparent:
                    render_mask = RenderMask.transparent
                else:
                    render_mask = RenderMask.opaque
        return {
            "indices": (wobject.geometry.positions.nitems, 1),
            "render_mask": render_mask,
        }

    def get_code(self):
        return (
            self.code_definitions()
            + self.code_common()
            + self.code_vertex()
            + self.code_fragment()
        )

    def code_vertex(self):
        return """
        struct VertexInput {
            @builtin(vertex_index) index : u32,
        };

        @vertex
        fn vs_main(in: VertexInput) -> Varyings {

            let i0 = i32(in.index);

            let raw_pos = load_s_positions(i0);
            let wpos = u_wobject.world_transform * vec4<f32>(raw_pos.xyz, 1.0);
            let npos = u_stdinfo.projection_transform * u_stdinfo.cam_transform * wpos;

            var varyings: Varyings;
            varyings.position = vec4<f32>(npos);
            varyings.world_pos = vec3<f32>(ndc_to_world_pos(npos));

            // Per-vertex colors
            $$ if vertex_color_channels == 1
            let cvalue = load_s_colors(i0);
            varyings.color = vec4<f32>(cvalue, cvalue, cvalue, 1.0);
            $$ elif vertex_color_channels == 2
            let cvalue = load_s_colors(i0);
            varyings.color = vec4<f32>(cvalue.r, cvalue.r, cvalue.r, cvalue.g);
            $$ elif vertex_color_channels == 3
            varyings.color = vec4<f32>(load_s_colors(i0), 1.0);
            $$ elif vertex_color_channels == 4
            varyings.color = vec4<f32>(load_s_colors(i0));
            $$ endif

            // Set texture coords
            $$ if colormap_dim == '1d'
            varyings.texcoord = f32(load_s_texcoords(i0));
            $$ elif colormap_dim == '2d'
            varyings.texcoord = vec2<f32>(load_s_texcoords(i0));
            $$ elif colormap_dim == '3d'
            varyings.texcoord = vec3<f32>(load_s_texcoords(i0));
            $$ endif

            return varyings;
        }
        """

    def code_fragment(self):
        return """
        @fragment
        fn fs_main(varyings: Varyings) -> FragmentOutput {

            $$ if color_mode == 'vertex'
                let color = varyings.color;
            $$ elif color_mode == 'map'
                let color = sample_colormap(varyings.texcoord);
            $$ else
                let color = u_material.color;
            $$ endif

            let physical_color = srgb2physical(color.rgb);
            let opacity = color.a * u_material.opacity;
            let out_color = vec4<f32>(physical_color, opacity);

            apply_clipping_planes(varyings.world_pos);
            return get_fragment_output(varyings.position.z, out_color);
        }
        """


@register_wgpu_render_function(Line, LineThinSegmentMaterial)
class ThinLineSegmentShader(ThinLineShader):
    def get_pipeline_info(self, wobject, shared):
        return {
            "primitive_topology": wgpu.PrimitiveTopology.line_list,
            "cull_mode": wgpu.CullMode.none,
        }



"""Predefined Colormaps.

The following colormaps are currently available:

.. currentmodule:: pygfx.utils.cm

.. autosummary::

    viridis
    plasma
    inferno
    magma
    cividis

"""

from ._cmdata_mpl import (
    _viridis_data,
    _plasma_data,
    _inferno_data,
    _magma_data,
    _cividis_data,
)

from ..resources import Texture


__all__ = ["viridis", "plasma", "inferno", "magma", "cividis"]


viridis = Texture(_viridis_data, dim=1).get_view()
plasma = Texture(_plasma_data, dim=1).get_view()
inferno = Texture(_inferno_data, dim=1).get_view()
magma = Texture(_magma_data, dim=1).get_view()
cividis = Texture(_cividis_data, dim=1).get_view()



""" Objects to view a scene.

.. autosummary::
    :toctree: cameras/

    Camera
    NDCCamera
    ScreenCoordsCamera
    PerspectiveCamera
    OrthographicCamera

"""

# flake8: noqa

from ._base import Camera, NDCCamera, ScreenCoordsCamera
from ._perspective import PerspectiveCamera
from ._orthographic import OrthographicCamera
